import pytest
from bunnyhopapi.request import RequestParser
import re


class TestRequestParser:
    @pytest.fixture
    def request_parser(self):
        routes = {}
        routes_with_params = {}
        return RequestParser(routes, routes_with_params)

    @pytest.mark.asyncio
    async def test_parse_request_valid_get(self, request_parser):
        request_data = (
            b"GET /test HTTP/1.1\r\nHost: localhost\r\nContent-Length: 0\r\n\r\n"
        )
        method, path, headers, body, query_params = await request_parser.parse_request(
            request_data
        )

        assert method == "GET"
        assert path == "/test"
        assert headers["Host"] == "localhost"
        assert body is None
        assert query_params == {}

    @pytest.mark.asyncio
    async def test_parse_request_with_query_params(self, request_parser):
        request_data = (
            b"GET /test?param1=value1&param2=value2 HTTP/1.1\r\nHost: localhost\r\n\r\n"
        )
        method, path, headers, body, query_params = await request_parser.parse_request(
            request_data
        )

        assert method == "GET"
        assert path == "/test"
        assert query_params == {"param1": "value1", "param2": "value2"}
        assert headers["Host"] == "localhost"
        assert body is None

    @pytest.mark.asyncio
    async def test_parse_request_with_body(self, request_parser):
        expected_body = '{"key": "value"}'
        request_data = (
            b"POST /test HTTP/1.1\r\n"
            b"Host: localhost\r\n"
            b"Content-Type: application/json\r\n"
            b"Content-Length: " + str(len(expected_body)).encode("utf-8") + b"\r\n\r\n"
            b"" + expected_body.encode("utf-8")
        )
        method, path, headers, body, query_params = await request_parser.parse_request(
            request_data
        )

        assert method == "POST"
        assert path == "/test"
        assert headers["Content-Type"] == "application/json"
        assert body == expected_body
        assert query_params == {}

    @pytest.mark.asyncio
    async def test_parse_request_invalid_format(self, request_parser):
        request_data = b"INVALID REQUEST"
        method, path, headers, body, query_params = await request_parser.parse_request(
            request_data
        )

        assert method is None
        assert path is None
        assert headers is None
        assert body is None
        assert query_params is None

    @pytest.mark.asyncio
    async def test_parse_request_missing_headers(self, request_parser):
        request_data = b"GET /test HTTP/1.1\r\n\r\n"
        method, path, headers, body, query_params = await request_parser.parse_request(
            request_data
        )

        assert method == "GET"
        assert path == "/test"
        assert headers == {}
        assert body is None
        assert query_params == {}

    @pytest.mark.asyncio
    async def test_parse_request_first_line_too_short(self, request_parser):
        request_data = b"GET\r\nHost: localhost\r\n\r\n"
        method, path, headers, body, query_params = await request_parser.parse_request(
            request_data
        )

        assert method is None
        assert path is None
        assert headers is None
        assert body is None
        assert query_params is None

    @pytest.mark.asyncio
    async def test_parse_request_empty_header_line(self, request_parser):
        request_data = (
            b"GET /test HTTP/1.1\r\nHost: localhost\n\r\n\r\n\r\n\r\n\r\n\r\n\r\n\r\n"
        )
        method, path, headers, body, query_params = await request_parser.parse_request(
            request_data
        )

        assert method == "GET"
        assert path == "/test"
        assert headers == {"Host": "localhost"}
        assert body is None
        assert query_params == {}

    @pytest.mark.asyncio
    async def test_parse_request_binary_body(self, request_parser):
        binary_body = b"\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00"
        request_data = (
            b"POST /upload HTTP/1.1\r\n"
            b"Host: localhost\r\n"
            b"Content-Type: application/octet-stream\r\n"
            b"Content-Length: "
            + str(len(binary_body)).encode("utf-8")
            + b"\r\n\r\n"
            + binary_body
        )
        method, path, headers, body, query_params = await request_parser.parse_request(
            request_data
        )

        assert method == "POST"
        assert path == "/upload"
        assert headers["Content-Type"] == "application/octet-stream"
        assert body == binary_body
        assert query_params == {}

    @pytest.mark.asyncio
    async def test_parse_request_value_error_in_body(self, request_parser):
        request_data = (
            b"POST /test HTTP/1.1\r\n"
            b"Host: localhost\r\n"
            b"Content-Type: application/json\r\n"
            b"Content-Length: invalid\r\n\r\n"
            b'{"key": "value"}'
        )
        method, path, headers, body, query_params = await request_parser.parse_request(
            request_data
        )

        assert method == "POST"
        assert path == "/test"
        assert headers["Content-Type"] == "application/json"
        assert body is None
        assert query_params == {}

    @pytest.mark.asyncio
    async def test_parse_request_generic_exception(self, request_parser):
        class MockBytes:
            def find(self, _):
                raise Exception("Mocked exception")

        request_data = MockBytes()
        method, path, headers, body, query_params = await request_parser.parse_request(
            request_data
        )

        assert method is None
        assert path is None
        assert headers is None
        assert body is None
        assert query_params is None

    def test_extract_params_route_not_in_routes_with_params(self, request_parser):
        path = "/test/123"
        route_path = "/test/<id>"
        result = request_parser._extract_params(path, route_path)
        assert result is None

    def test_extract_params_no_match(self, request_parser):
        request_parser.routes_with_params = {
            "/test/<id>": re.compile(r"^/test/(?P<id>\d+)$")
        }
        path = "/test/abc"
        route_path = "/test/<id>"
        result = request_parser._extract_params(path, route_path)
        assert result is None

    def test_extract_params_valid_match(self, request_parser):
        request_parser.routes_with_params = {
            "/test/<id>": re.compile(r"^/test/(?P<id>\d+)$")
        }
        path = "/test/123"
        route_path = "/test/<id>"
        result = request_parser._extract_params(path, route_path)
        assert result == {"id": "123"}



import re
import inspect
from typing import get_type_hints, get_origin
from pydantic import BaseModel
from bunnyhopapi.models import PathParam, QueryParam

SWAGGER_JSON = {
    "openapi": "3.0.0",
    "info": {
        "title": "BunnyHop API",
        "version": "1.0.0",
    },
    "paths": {},
}

TYPE_MAPPING = {
    "int": "integer",
    "float": "number",
    "str": "string",
    "bool": "boolean",
}

SWAGGER_JSON["components"] = SWAGGER_JSON.get("components", {})
SWAGGER_JSON["components"]["schemas"] = SWAGGER_JSON["components"].get("schemas", {})
SWAGGER_JSON["components"]["securitySchemes"] = {
    "Authorization": {"type": "apiKey", "name": "Authorization", "in": "header"}
}

SWAGGER_JSON["security"] = [{"Authorization": []}]


class SwaggerGenerator:
    @staticmethod
    def generate_path_item(path: str, methods: dict):
        swagger_path = re.sub(r"<(\w+)>", r"{\1}", path)

        if not swagger_path.startswith("/"):
            swagger_path = "/" + swagger_path

        swagger_path = re.sub(r"/+", "/", swagger_path)

        SWAGGER_JSON["paths"][swagger_path] = {}

        for method, details in methods.items():
            handler = details.get("handler")
            if not callable(handler):
                raise TypeError(f"{handler} is not a callable object.")

            type_hints = get_type_hints(handler)

            operation = SwaggerGenerator._generate_operation(
                method, path, type_hints, details, handler
            )
            SWAGGER_JSON["paths"][swagger_path][method.lower()] = operation

    @staticmethod
    def _generate_operation(
        method: str, path: str, type_hints: dict, details: dict, handler
    ):
        response_schema = {}
        parameters = []
        request_body = None

        parameters.extend(SwaggerGenerator._process_path_params(type_hints, handler))
        request_body = SwaggerGenerator._process_body_params(type_hints)
        response_schema = SwaggerGenerator._process_response_types(type_hints)

        operation = {
            "summary": f"Handler for {method} {path}",
            "responses": response_schema
            or {
                200: {
                    "description": "Successful response",
                    "content": {"application/json": {"schema": {}}},
                }
            },
        }

        if parameters:
            operation["parameters"] = parameters

        if request_body and method.upper() in ["POST", "PUT", "PATCH"]:
            operation["requestBody"] = request_body

        return operation

    @staticmethod
    def _process_path_params(type_hints: dict, handler):
        parameters = []
        for param_name, param_type in type_hints.items():
            if get_origin(param_type) is PathParam:
                type_name = param_type.__args__[0].__name__
                swagger_type = TYPE_MAPPING.get(type_name, type_name)
                parameters.append(
                    {
                        "name": param_name,
                        "in": "path",
                        "required": True,
                        "schema": {"type": swagger_type},
                    }
                )
            elif get_origin(param_type) is QueryParam:
                type_name = param_type.__args__[0].__name__
                swagger_type = TYPE_MAPPING.get(type_name, type_name)

                signature = inspect.signature(handler)
                parameter = signature.parameters.get(param_name)
                is_required = not bool(
                    parameter is not None
                    and parameter.default is not inspect.Parameter.empty
                )

                parameters.append(
                    {
                        "name": param_name,
                        "in": "query",
                        "required": is_required,
                        "schema": {
                            "type": swagger_type,
                        },
                    }
                )
        return parameters

    @staticmethod
    def _process_body_params(type_hints: dict):
        for param_name, param_type in type_hints.items():
            if (
                inspect.isclass(param_type)
                and issubclass(param_type, BaseModel)
                and not isinstance(param_type, PathParam)
            ):
                if "components" not in SWAGGER_JSON:
                    SWAGGER_JSON["components"] = {"schemas": {}}
                SWAGGER_JSON["components"]["schemas"][param_type.__name__] = (
                    param_type.model_json_schema()
                )

                return {
                    "required": True,
                    "content": {
                        "application/json": {
                            "schema": param_type.model_json_schema(
                                ref_template="#/components/schemas/{model}"
                            )
                        }
                    },
                }
        return None

    @staticmethod
    def _process_response_types(type_hints: dict):
        response_schema = {}

        if "return" in type_hints:
            return_type = type_hints["return"]

            if isinstance(return_type, dict):
                for status_code, model in return_type.items():
                    if inspect.isclass(model) and issubclass(model, BaseModel):
                        response_schema.update(
                            SwaggerGenerator._add_response_model(status_code, model)
                        )
                    elif model is None:
                        response_schema[status_code] = {
                            "description": f"Response with status {status_code}"
                        }
            elif inspect.isclass(return_type) and issubclass(return_type, BaseModel):
                response_schema.update(
                    SwaggerGenerator._add_response_model(200, return_type)
                )

        return response_schema

    @staticmethod
    def _add_response_model(status_code, model):
        if "components" not in SWAGGER_JSON:
            SWAGGER_JSON["components"] = {"schemas": {}}
        if model.__name__ not in SWAGGER_JSON["components"]["schemas"]:
            SWAGGER_JSON["components"]["schemas"][model.__name__] = (
                model.model_json_schema(ref_template="#/components/schemas/{model}")
            )

        return {
            status_code: {
                "description": f"Response with status {status_code}",
                "content": {
                    "application/json": {
                        "schema": {"$ref": f"#/components/schemas/{model.__name__}"}
                    }
                },
            }
        }

    @staticmethod
    def get_swagger_ui_html():
        return """
        <!DOCTYPE html>
        <html>
        <head>
            <title>Swagger UI</title>
            <link rel="stylesheet" type="text/css" href="https://cdnjs.cloudflare.com/ajax/libs/swagger-ui/5.20.0/swagger-ui.css">
        </head>
        <body>
            <div id="swagger-ui"></div>
            <script src="https://cdnjs.cloudflare.com/ajax/libs/swagger-ui/5.20.0/swagger-ui-bundle.js"></script>
            <script>
                const ui = SwaggerUIBundle({
                    url: '/swagger.json',
                    dom_id: '#swagger-ui',
                });
            </script>
        </body>
        </html>
        """



import pytest
from unittest.mock import AsyncMock, Mock, patch, ANY
import struct
from bunnyhopapi.websocket import WebSocketHandler


@pytest.fixture
def websocket_handler():
    handlers = {
        "/ws/test": {
            "handler": AsyncMock(return_value="response"),
            "middleware": None,
            "connection": AsyncMock(return_value=True),
            "disconnect": AsyncMock(),
        }
    }
    return WebSocketHandler(handlers)


@pytest.mark.asyncio
class TestWebSocketHandler:
    async def test_handle_websocket_connection_success(self, websocket_handler):
        reader = AsyncMock()
        writer = AsyncMock()
        writer.write = Mock()
        writer.close = Mock()
        headers = {"Sec-WebSocket-Key": "dGhlIHNhbXBsZSBub25jZQ=="}

        await websocket_handler.handle_websocket(reader, writer, "/ws/test", headers)

        websocket_handler.websocket_handlers["/ws/test"][
            "connection"
        ].assert_called_once()
        writer.write.assert_called()
        writer.drain.assert_called()

    async def test_handle_websocket_no_handler(self, websocket_handler):
        reader = AsyncMock()
        writer = AsyncMock()
        writer.write = Mock()
        headers = {"Sec-WebSocket-Key": "dGhlIHNhbXBsZSBub25jZQ=="}

        await websocket_handler.handle_websocket(reader, writer, "/ws/unknown", headers)

        writer.write.assert_not_called()

    async def test_handle_websocket_no_key(self, websocket_handler):
        reader = AsyncMock()
        writer = AsyncMock()
        writer.write = AsyncMock()
        headers = {}

        await websocket_handler.handle_websocket(reader, writer, "/ws/test", headers)

        websocket_handler.websocket_handlers["/ws/test"][
            "connection"
        ].assert_not_called()
        writer.write.assert_not_called()

    async def test_handle_websocket_message_processing(self, websocket_handler):
        reader = AsyncMock()
        writer = AsyncMock()
        writer.write = Mock()
        writer.close = Mock()
        headers = {"Sec-WebSocket-Key": "dGhlIHNhbXBsZSBub25jZQ=="}
        reader.read.side_effect = [
            b"\x81\x05",  # Frame header
            b"hello",  # Frame payload
            b"\x88\x00",  # Close frame
        ]

        mock_uuid = "mocked-uuid"
        with patch("uuid.uuid4", return_value=mock_uuid):
            await websocket_handler.handle_websocket(
                reader, writer, "/ws/test", headers
            )

        websocket_handler.websocket_handlers["/ws/test"][
            "handler"
        ].assert_called_once_with(
            connection_id=mock_uuid, message="hello", headers=headers
        )
        writer.write.assert_called()

    async def test_handle_websocket_disconnect(self, websocket_handler):
        reader = AsyncMock()
        writer = AsyncMock()
        writer.write = Mock()
        writer.close = Mock()
        headers = {"Sec-WebSocket-Key": "dGhlIHNhbXBsZSBub25jZQ=="}
        reader.read.side_effect = [
            b"\x88\x00",  # Close frame
        ]

        mock_uuid = "mocked-uuid"
        with patch("uuid.uuid4", return_value=mock_uuid):
            await websocket_handler.handle_websocket(
                reader, writer, "/ws/test", headers
            )

        websocket_handler.websocket_handlers["/ws/test"][
            "disconnect"
        ].assert_called_once_with(connection_id=mock_uuid, headers=headers)
        writer.close.assert_called()

    async def test_handle_websocket_payload_length_126(self, websocket_handler):
        reader = AsyncMock()
        writer = AsyncMock()
        writer.write = Mock()
        writer.close = Mock()
        headers = {"Sec-WebSocket-Key": "dGhlIHNhbXBsZSBub25jZQ=="}
        reader.read.side_effect = [
            b"\x81\x7e",  # Frame header with payload length 126
            struct.pack(">H", 5),  # Extended payload length
            b"hello",  # Frame payload
            b"\x88\x00",  # Close frame
        ]

        await websocket_handler.handle_websocket(reader, writer, "/ws/test", headers)

        websocket_handler.websocket_handlers["/ws/test"][
            "handler"
        ].assert_called_once_with(connection_id=ANY, message="hello", headers=headers)
        writer.write.assert_called()

    async def test_handle_websocket_payload_length_127(self, websocket_handler):
        reader = AsyncMock()
        writer = AsyncMock()
        writer.write = Mock()
        writer.close = Mock()
        headers = {"Sec-WebSocket-Key": "dGhlIHNhbXBsZSBub25jZQ=="}
        reader.read.side_effect = [
            b"\x81\x7f",  # Frame header with payload length 127
            struct.pack(">Q", 5),  # Extended payload length
            b"hello",  # Frame payload
            b"\x88\x00",  # Close frame
        ]

        await websocket_handler.handle_websocket(reader, writer, "/ws/test", headers)

        websocket_handler.websocket_handlers["/ws/test"][
            "handler"
        ].assert_called_once_with(connection_id=ANY, message="hello", headers=headers)
        writer.write.assert_called()

    async def test_handle_websocket_no_mask(self, websocket_handler):
        reader = AsyncMock()
        writer = AsyncMock()
        writer.write = Mock()
        writer.close = Mock()
        headers = {"Sec-WebSocket-Key": "dGhlIHNhbXBsZSBub25jZQ=="}
        reader.read.side_effect = [
            b"\x81\x05",  # Frame header without masking
            b"hello",  # Frame payload
            b"\x88\x00",  # Close frame
        ]

        await websocket_handler.handle_websocket(reader, writer, "/ws/test", headers)

        websocket_handler.websocket_handlers["/ws/test"][
            "handler"
        ].assert_called_once_with(connection_id=ANY, message="hello", headers=headers)
        writer.write.assert_called()

    async def test_handle_websocket_no_sec_websocket_key(self, websocket_handler):
        reader = AsyncMock()
        writer = AsyncMock()
        writer.write = Mock()
        headers = {}  # Missing Sec-WebSocket-Key

        await websocket_handler.handle_websocket(reader, writer, "/ws/test", headers)

        websocket_handler.websocket_handlers["/ws/test"][
            "connection"
        ].assert_not_called()
        writer.write.assert_not_called()

    async def test_handle_websocket_connection_auth_failure(self, websocket_handler):
        reader = AsyncMock()
        writer = AsyncMock()
        writer.write = Mock()
        writer.close = Mock()
        headers = {"Sec-WebSocket-Key": "dGhlIHNhbXBsZSBub25jZQ=="}
        websocket_handler.websocket_handlers["/ws/test"][
            "connection"
        ].return_value = False

        await websocket_handler.handle_websocket(reader, writer, "/ws/test", headers)

        writer.close.assert_called_once()
        websocket_handler.websocket_handlers["/ws/test"]["handler"].assert_not_called()

    async def test_handle_websocket_masked_payload(self, websocket_handler):
        reader = AsyncMock()
        writer = AsyncMock()
        writer.write = Mock()
        writer.close = Mock()
        headers = {"Sec-WebSocket-Key": "dGhlIHNhbXBsZSBub25jZQ=="}

        # Masking key: [0x37, 0xfa, 0x21, 0x3d]
        # Original payload: "hello" -> [0x68, 0x65, 0x6c, 0x6c, 0x6f]
        # Masked payload: [0x5f, 0x9f, 0x4d, 0x51, 0x58]
        reader.read.side_effect = [
            b"\x81\x85",  # Frame header with masking and payload length 5
            b"\x37\xfa\x21\x3d",  # Masking key
            b"\x5f\x9f\x4d\x51\x58",  # Masked payload
            b"\x88\x00",  # Close frame
        ]

        await websocket_handler.handle_websocket(reader, writer, "/ws/test", headers)

        websocket_handler.websocket_handlers["/ws/test"][
            "handler"
        ].assert_called_once_with(connection_id=ANY, message="hello", headers=headers)
        writer.write.assert_called()

    async def test_handle_websocket_async_generator_response(self, websocket_handler):
        async def async_generator_response(*args, **kwargs):
            yield "response1"
            yield "response2"

        websocket_handler.websocket_handlers["/ws/test"]["handler"] = (
            async_generator_response
        )

        reader = AsyncMock()
        writer = AsyncMock()
        writer.write = Mock()
        writer.close = Mock()
        headers = {"Sec-WebSocket-Key": "dGhlIHNhbXBsZSBub25jZQ=="}
        reader.read.side_effect = [
            b"\x81\x05",  # Frame header
            b"hello",  # Frame payload
            b"\x88\x00",  # Close frame
        ]

        await websocket_handler.handle_websocket(reader, writer, "/ws/test", headers)

        writer.write.assert_any_call(bytearray(b"\x81\tresponse1"))
        writer.write.assert_any_call(bytearray(b"\x81\tresponse2"))

    async def test_handle_websocket_with_middleware(self, websocket_handler):
        middleware_mock = AsyncMock(return_value="middleware_response")
        websocket_handler.websocket_handlers["/ws/test"]["middleware"] = middleware_mock

        reader = AsyncMock()
        writer = AsyncMock()
        writer.write = Mock()
        writer.close = Mock()
        headers = {"Sec-WebSocket-Key": "dGhlIHNhbXBsZSBub25jZQ=="}
        reader.read.side_effect = [
            b"\x81\x05",  # Frame header
            b"hello",  # Frame payload
            b"\x88\x00",  # Close frame
        ]

        mock_uuid = "mocked-uuid"
        with patch("uuid.uuid4", return_value=mock_uuid):
            await websocket_handler.handle_websocket(
                reader, writer, "/ws/test", headers
            )

        middleware_mock.assert_called_once_with(
            connection_id=mock_uuid, message="hello", headers=headers
        )
        writer.write.assert_called_with(bytearray(b"\x81\x13middleware_response"))



import inspect
from . import logger
from .request import RequestParser
from .response import ResponseHandler
from .handlers import RouteHandler
from .websocket import WebSocketHandler
from .models import RouterBase


class ClientHandler:
    def __init__(
        self,
        routes: dict,
        routes_with_params: dict,
        websocket_handlers: dict,
        cors: bool = False,
    ):
        self.request_parser = RequestParser(routes, routes_with_params)
        self.response_handler = ResponseHandler(cors)
        self.route_handler = RouteHandler(routes, routes_with_params)
        self.websocket_handler = WebSocketHandler(websocket_handlers)

    async def handle_client(self, reader, writer):
        request_data = await self._read_request(reader)
        if request_data is None:
            await self._send_error_response(writer, 400, "Invalid request")
            return

        (
            method,
            path,
            headers,
            body,
            query_params,
        ) = await self.request_parser.parse_request(request_data)

        if method is None:
            await self._send_error_response(writer, 400, "Invalid request")
            return

        if method.upper() == "OPTIONS":
            await self._handle_options(writer)
            return

        if (
            headers.get("Connection", "").lower() == "upgrade"
            and headers.get("Upgrade", "").lower() == "websocket"
        ):
            await self.websocket_handler.handle_websocket(reader, writer, path, headers)
            return

        response = await self.route_handler.execute_handler(
            path, method, body, headers, query_params
        )
        await self._send_response(writer, response)

    async def _read_request(self, reader) -> bytes | None:
        try:
            request_data = b""
            while True:
                chunk = await reader.read(8192)
                if not chunk:
                    break
                request_data += chunk
                if b"\r\n\r\n" in request_data:
                    break
            return request_data
        except Exception as e:
            logger.error(f"Error reading from client: {e}")
            return None

    async def _send_error_response(self, writer, status_code: int, message: str):
        response = self.response_handler.prepare_response(
            "application/json", status_code, {"error": message}
        )
        writer.write(response)
        await writer.drain()
        writer.close()
        await writer.wait_closed()

    async def _handle_options(self, writer):
        response = self.response_handler.prepare_options_response()
        writer.write(response)
        await writer.drain()
        writer.close()
        await writer.wait_closed()

    async def _send_response(self, writer, response):
        try:
            if (
                isinstance(response, dict)
                and response.get("content_type") == RouterBase.CONTENT_TYPE_SSE
                and inspect.isasyncgen(response.get("response_data"))
            ):
                prepared, generator = self.response_handler.prepare_response(
                    response["content_type"],
                    response["status_code"],
                    response["response_data"],
                )

                writer.write(prepared)
                await writer.drain()

                try:
                    async for chunk in generator:
                        if isinstance(chunk, str):
                            chunk = chunk.encode("utf-8")
                        writer.write(chunk)
                        await writer.drain()
                finally:
                    writer.close()
                    await writer.wait_closed()
                return

            if isinstance(response, tuple) and len(response) == 3:
                content_type, status_code, response_data = response
            elif isinstance(response, dict):
                content_type = response["content_type"]
                status_code = response["status_code"]
                response_data = response["response_data"]
            else:
                content_type = "application/json"
                status_code = 500
                response_data = {
                    "error": "Internal server error",
                    "message": "Unknown response format",
                }

            prepared = self.response_handler.prepare_response(
                content_type, status_code, response_data
            )

            writer.write(prepared)
            await writer.drain()

        except Exception as e:
            logger.error(f"Error sending response: {e}")
        finally:
            writer.close()
            await writer.wait_closed()



import pytest
from unittest.mock import AsyncMock, MagicMock, patch
from bunnyhopapi.server import Server
from bunnyhopapi.templates import serve_static_file


class TestServer:
    @pytest.fixture
    def server_instance(self):
        return Server(cors=True, port=8000)

    def test_add_swagger(self, server_instance):
        server_instance.add_swagger()
        assert "/swagger.json" in server_instance.routes
        assert "/docs" in server_instance.routes

    @pytest.mark.asyncio
    async def test_include_static_folder(self, server_instance, tmp_path):
        content_file = "Hello, world!"
        static_folder = tmp_path / "static"
        static_folder.mkdir()
        (static_folder / "test.txt").write_text(content_file)

        server_instance.include_static_folder(str(static_folder))

        route = "/static/test.txt"
        assert route in server_instance.routes
        handler = server_instance.routes[route]["GET"]["handler"]
        assert handler is not None

        response = await serve_static_file(static_folder / "test.txt")
        status_code, content = response
        assert status_code == 200
        assert content == content_file

    @pytest.mark.asyncio
    async def test_include_static_folder_not_exists(self, server_instance):
        non_existent_path = "/path/does/not/exist"

        with patch("bunnyhopapi.server.logger.error") as mock_logger:
            server_instance.include_static_folder(non_existent_path)

            mock_logger.assert_called_once_with(
                f"Static folder does not exist: {non_existent_path}"
            )

    @pytest.mark.asyncio
    async def test_include_static_folder_not_directory(self, server_instance, tmp_path):
        not_a_directory = tmp_path / "file.txt"
        not_a_directory.write_text("This is a file, not a directory.")

        with patch("bunnyhopapi.server.logger.error") as mock_logger:
            server_instance.include_static_folder(str(not_a_directory))

            mock_logger.assert_called_once_with(
                f"Provided path is not a directory: {not_a_directory}"
            )

    @pytest.mark.asyncio
    async def test_run_server(self):
        with patch(
            "bunnyhopapi.server.asyncio.start_server", new_callable=AsyncMock
        ) as mock_start_server:
            server_instance = Server(cors=True, port=8000)
            mock_server = AsyncMock()
            mock_start_server.return_value = mock_server

            with patch("bunnyhopapi.server.uvloop.install"):
                await server_instance._run()

            mock_start_server.assert_called_once()
            mock_server.serve_forever.assert_called_once()

    def test_run(self):
        with (
            patch("bunnyhopapi.server.uvloop.install") as mock_uvloop_install,
            patch("bunnyhopapi.server.Process") as mock_process,
            patch("bunnyhopapi.server.Server._run", return_value=None) as mock_run,
        ):
            mock_process_instance = MagicMock()
            mock_process.return_value = mock_process_instance

            server_instance = Server(cors=True, port=8000)
            server_instance.run(workers=2)

            mock_uvloop_install.assert_called_once()

            assert mock_process.call_count == 2
            mock_process_instance.start.assert_called()
            mock_process_instance.join.assert_called()

            for call_args in mock_process.call_args_list:
                assert "target" in call_args.kwargs
                assert callable(call_args.kwargs["target"])

    def test_run_keyboard_interrupt(self):
        with (
            patch("bunnyhopapi.server.uvloop.install") as mock_uvloop_install,
            patch("bunnyhopapi.server.Process") as mock_process,
            patch("bunnyhopapi.server.logger.info") as mock_logger,
        ):
            mock_process_instance = MagicMock()
            mock_process.return_value = mock_process_instance

            server_instance = Server(cors=True, port=8000)

            mock_process_instance.join.side_effect = KeyboardInterrupt

            server_instance.run(workers=2)

            mock_uvloop_install.assert_called_once()

            mock_logger.assert_any_call("Server stopped by user")

            mock_process_instance.terminate.assert_called()
            mock_logger.assert_any_call("Server stopped")

    def test_handle_exit(self):
        with patch("os._exit") as mock_exit:
            with patch("bunnyhopapi.server.logger.info") as mock_logger:
                from signal import SIGINT
                from bunnyhopapi.server import handle_exit

                handle_exit(SIGINT, None)
                mock_logger.assert_called_with("Shutting down server...")
                mock_exit.assert_called_once_with(0)

    @pytest.mark.asyncio
    async def test_generate_swagger_json(self, server_instance):
        server_instance.add_swagger()
        server_instance.add_route("/test", "GET", lambda: None)
        status_code, swagger_json = await server_instance.generate_swagger_json()

        assert status_code == 200

        assert "/test" in swagger_json["paths"]
        assert "/docs" not in swagger_json["paths"]
        assert "/swagger.json" not in swagger_json["paths"]

    @pytest.mark.asyncio
    async def test_swagger_ui_handler(self, server_instance):
        status_code, response = await server_instance.swagger_ui_handler()

        assert status_code == 200

        assert "<!DOCTYPE html>" in response
        assert "<title>Swagger UI</title>" in response

    @pytest.mark.asyncio
    async def test_start_worker_run(self):
        with patch(
            "bunnyhopapi.server.Server._run", new_callable=AsyncMock
        ) as mock_run:
            server_instance = Server(cors=True, port=8000)
            server_instance._start_worker()

            mock_run.assert_called_once()

    def test_start_worker_keyboard_interrupt(self):
        with (
            patch(
                "bunnyhopapi.server.Server._run", side_effect=KeyboardInterrupt
            ) as mock_run,
            patch("bunnyhopapi.server.logger.info") as mock_logger,
        ):
            server_instance = Server(cors=True, port=8000)
            server_instance._start_worker()

            mock_run.assert_called_once()
            mock_logger.assert_called_with("Worker received stop signal")



import pytest
from unittest.mock import Mock
from bunnyhopapi.handlers import RouteHandler
from bunnyhopapi.models import PathParam
from pydantic import BaseModel
import inspect


class TestRouteHandler:
    @pytest.fixture
    def route_handler(self):
        return RouteHandler()

    @pytest.mark.asyncio
    async def test_execute_handler_route_not_found(self, route_handler):
        route_handler._find_route = Mock(return_value=None)
        response = await route_handler.execute_handler("/invalid", "GET")
        assert response == (
            "application/json",
            404,
            {"error": "Route GET /invalid not found"},
        )

    @pytest.mark.asyncio
    async def test_execute_handler_success(self, route_handler):
        async def mock_handler(headers, **kwargs):
            return 200, {"message": "Success"}

        route_handler._find_route = Mock(
            return_value=(
                {"handler": mock_handler, "content_type": "application/json"},
                {},
            )
        )
        response = await route_handler.execute_handler("/valid", "GET", query_params={})
        assert response["status_code"] == 200
        assert response["response_data"] == {"message": "Success"}

    def test_find_route_exact_match(self, route_handler):
        route_handler.routes = {"/test": {"GET": {"handler": Mock()}}}
        route, params = route_handler._find_route("/test", "GET")
        assert route is not None
        assert params == {}

    def test_find_route_method_not_in_methods(self, route_handler):
        route_handler.routes = {"/test": {"POST": {"handler": Mock()}}}
        result = route_handler._find_route("/test", "GET")
        assert result is None

    def test_find_route_with_params(self, route_handler):
        route_handler.routes = {"/test/<id>": {"GET": {"handler": Mock()}}}
        route_handler.routes_with_params = {"/test/<id>": Mock()}
        route_handler._extract_params = Mock(return_value={"id": "123"})
        route, params = route_handler._find_route("/test/123", "GET")
        assert route is not None
        assert params == {"id": "123"}

    def test_find_route_no_match(self, route_handler):
        route_handler.routes = {"/test/<id>": {"GET": {"handler": Mock()}}}
        route_handler.routes_with_params = {"/test/<id>": Mock()}
        route_handler._extract_params = Mock(return_value=None)
        result = route_handler._find_route("/test/abc", "GET")
        assert result is None

    def test_find_route_with_params_found(self, route_handler):
        route_handler.routes = {"/test/<id>": {"GET": {"handler": Mock()}}}
        route_handler.routes_with_params = {"/test/<id>": Mock()}
        route_handler._extract_params = Mock(return_value={"id": "123"})

        route, params = route_handler._find_route("/test/123", "GET")
        assert route is not None
        assert params == {"id": "123"}

    def test_validate_params_success(self, route_handler):
        type_hints = {"id": PathParam[int]}
        params = {"id": "123"}
        validated = route_handler._validate_params(params, type_hints, {})
        assert validated["id"] == 123

    def test_validate_params_invalid(self, route_handler):
        type_hints = {"id": PathParam[int]}
        params = {"id": "abc"}
        with pytest.raises(ValueError):
            route_handler._validate_params(params, type_hints, {})

    def test_validate_params_param_not_in_type_hints(self, route_handler):
        type_hints = {"id": PathParam[int]}
        params = {"name": "test"}
        validated = route_handler._validate_params(params, type_hints, {})
        assert "name" not in validated

    def test_validate_params_assign_param_value(self, route_handler):
        type_hints = {"id": PathParam[int], "name": str}
        params = {"name": "test"}
        validated = route_handler._validate_params(params, type_hints, {})
        assert validated["name"] == "test"

    def test_validate_body_success(self, route_handler):
        class BodyModel(BaseModel):
            name: str

        type_hints = {"body": BodyModel}
        body = '{"name": "test"}'
        result = route_handler._validate_body(body, type_hints)
        assert result["status_code"] == 200
        assert result["validated_params"]["body"].name == "test"

    def test_validate_body_invalid_json(self, route_handler):
        class BodyModel(BaseModel):
            name: str

        type_hints = {"body": BodyModel}
        body = '{"name": "test"'
        result = route_handler._validate_body(body, type_hints)
        assert result["status_code"] == 400
        assert "Invalid JSON format" in result["error"]["error"]

    def test_validate_body_validation_error(self, route_handler):
        class BodyModel(BaseModel):
            name: str

        type_hints = {"body": BodyModel}
        body = '{"age": 30}'
        result = route_handler._validate_body(body, type_hints)
        assert result["status_code"] == 422
        assert "Validation error" in result["error"]["error"]

    def test_extract_params_route_not_in_routes_with_params(self, route_handler):
        route_handler.routes_with_params = {}
        result = route_handler._extract_params("/test/123", "/test/<id>")
        assert result is None

    def test_extract_params_no_match(self, route_handler):
        mock_pattern = Mock()
        mock_pattern.match.return_value = None
        route_handler.routes_with_params = {"/test/<id>": mock_pattern}
        result = route_handler._extract_params("/test/abc", "/test/<id>")
        assert result is None

    def test_extract_params_match(self, route_handler):
        mock_match = Mock()
        mock_match.groupdict.return_value = {"id": "123"}
        mock_pattern = Mock()
        mock_pattern.match.return_value = mock_match
        route_handler.routes_with_params = {"/test/<id>": mock_pattern}
        result = route_handler._extract_params("/test/123", "/test/<id>")
        assert result == {"id": "123"}

    @pytest.mark.asyncio
    async def test_process_handler_validation_error(self, route_handler):
        async def mock_handler(headers, **kwargs):
            return 200, {"message": "Success"}

        route_info = {
            "handler": mock_handler,
            "content_type": "application/json",
            "params": {"id": "invalid"},
        }

        route_handler._validate_params = Mock(
            side_effect=ValueError("Invalid parameter")
        )

        response = await route_handler._process_handler(
            route_info, "/test/123", "GET", headers={}
        )

        assert response["content_type"] == "application/json"
        assert response["status_code"] == 422
        assert response["response_data"] == {"error": "Invalid parameter"}

    @pytest.mark.asyncio
    async def test_process_handler_body_validation_error(self, route_handler):
        async def mock_handler(headers, **kwargs):
            return 200, {"message": "Success"}

        route_info = {
            "handler": mock_handler,
            "content_type": "application/json",
        }

        route_handler._validate_body = Mock(
            return_value={
                "status_code": 400,
                "error": {"error": "Invalid JSON format"},
                "validated_params": {},
            }
        )

        response = await route_handler._process_handler(
            route_info, "/test", "POST", body="invalid_json", headers={}
        )

        assert response["content_type"] == "application/json"
        assert response["status_code"] == 400
        assert response["response_data"] == {"error": "Invalid JSON format"}

    @pytest.mark.asyncio
    async def test_process_handler_body_validation_success(self, route_handler):
        async def mock_handler(headers, **kwargs):
            return 200, {"message": "Success"}

        route_info = {
            "handler": mock_handler,
            "content_type": "application/json",
        }

        route_handler._validate_body = Mock(
            return_value={
                "status_code": 200,
                "error": None,
                "validated_params": {"body": {"key": "value"}},
            }
        )

        route_handler._validate_params = Mock(return_value={})

        response = await route_handler._process_handler(
            route_info, "/test", "POST", body='{"key": "value"}', headers={}
        )

        assert response["content_type"] == "application/json"
        assert response["status_code"] == 200
        assert response["response_data"] == {"message": "Success"}

    @pytest.mark.asyncio
    async def test_process_handler_async_generator(self, route_handler):
        async def mock_handler(headers, **kwargs):
            yield "data"

        route_info = {
            "handler": mock_handler,
            "content_type": "text/event-stream",
        }

        route_handler._validate_params = Mock(return_value={})

        response = await route_handler._process_handler(
            route_info, "/test", "GET", headers={}
        )

        assert response["content_type"] == "text/event-stream"
        assert response["status_code"] == 200
        assert inspect.isasyncgen(response["response_data"])

    @pytest.mark.asyncio
    async def test_process_handler_invalid_response_tuple(self, route_handler):
        async def mock_handler(headers, **kwargs):
            return 200  # Invalid response, not a tuple

        route_info = {
            "handler": mock_handler,
            "content_type": "application/json",
        }

        response = await route_handler._process_handler(
            route_info, "/test", "GET", headers={}
        )

        assert response["content_type"] == "application/json"
        assert response["status_code"] == 500
        assert response["response_data"]["error"] == "Internal server error"
        assert (
            "Handler must return a tuple of (status_code, response_data)"
            in response["response_data"]["message"]
        )

    @pytest.mark.asyncio
    async def test_process_handler_response_model_validation_success(
        self, route_handler
    ):
        class ResponseModel(BaseModel):
            message: str

        async def mock_handler(headers, **kwargs):
            return 200, {"message": "Success"}

        route_info = {
            "handler": mock_handler,
            "content_type": "application/json",
        }

        route_handler._validate_params = Mock(return_value={})
        route_handler._validate_body = Mock(
            return_value={"status_code": 200, "error": None, "validated_params": {}}
        )
        route_info["return"] = {200: ResponseModel}

        response = await route_handler._process_handler(
            route_info, "/test", "GET", headers={}
        )

        assert response["content_type"] == "application/json"
        assert response["status_code"] == 200
        assert response["response_data"] == {"message": "Success"}

    @pytest.mark.asyncio
    async def test_process_handler_response_model_validation_error(self, route_handler):
        class ResponseModel(BaseModel):
            message: str

        async def mock_handler(headers, **kwargs) -> {200: ResponseModel}:
            return 200, {"invalid_field": "Error"}

        route_info = {
            "handler": mock_handler,
            "content_type": "application/json",
            "return": {200: ResponseModel},
        }

        route_handler._validate_params = Mock(return_value={})
        route_handler._validate_body = Mock(
            return_value={"status_code": 200, "error": None, "validated_params": {}}
        )

        response = await route_handler._process_handler(
            route_info, "/test", "GET", headers={}
        )

        assert response["content_type"] == "application/json"
        assert response["status_code"] == 422
        assert response["response_data"]["error"] == "Validation error"
        assert "details" in response["response_data"]



class RequestParser:
    def __init__(self, routes: dict, routes_with_params: dict):
        self.routes = routes
        self.routes_with_params = routes_with_params

    def _extract_params(self, path: str, route_path: str):
        if route_path not in self.routes_with_params:
            return None

        compiled_pattern = self.routes_with_params[route_path]
        match = compiled_pattern.match(path)
        if not match:
            return None

        return match.groupdict()

    async def parse_request(self, request_data: bytes):
        try:
            header_end = request_data.find(b"\r\n\r\n")
            if header_end == -1:
                return None, None, None, None, None

            header_data = request_data[:header_end]

            headers_text = header_data.decode("latin-1")

            first_line_end = headers_text.find("\r\n")
            if first_line_end == -1:
                first_line = headers_text.split(" ", 2)
            else:
                first_line = headers_text[:first_line_end].split(" ", 2)

            if len(first_line) < 2:
                return None, None, None, None, None

            method = first_line[0]
            raw_path = first_line[1]

            path_end = raw_path.find("?")
            if path_end == -1:
                path = raw_path
                query_params = {}
            else:
                path = raw_path[:path_end]
                query_string = raw_path[path_end + 1 :]
                query_params = {}

                if query_string:
                    for pair in query_string.split("&"):
                        if "=" in pair:
                            key, value = pair.split("=", 1)
                            query_params[key] = value.split(",")[0]

            headers = {}
            if first_line_end != -1:
                header_lines = headers_text[first_line_end + 2 :].split("\r\n")
                for line in header_lines:
                    colon_pos = line.find(":")
                    if colon_pos > 0:
                        key = line[:colon_pos].strip()
                        value = line[colon_pos + 1 :].strip()
                        headers[key] = value

            body = None
            if "Content-Length" in headers:
                try:
                    content_length = int(headers["Content-Length"])
                    body_start = header_end + 4
                    if content_length > 0 and (body_start + content_length) <= len(
                        request_data
                    ):
                        if (
                            headers.get("Content-Type", "").startswith("text/")
                            or headers.get("Content-Type", "") == "application/json"
                        ):
                            body = request_data[
                                body_start : body_start + content_length
                            ].decode("utf-8")
                        else:
                            body = request_data[
                                body_start : body_start + content_length
                            ]
                except (ValueError, UnicodeDecodeError):
                    pass

            return method, path, headers, body, query_params

        except Exception as e:
            return None, None, None, None, None



import pytest
from bunnyhopapi.models import RouterBase
from bunnyhopapi.response import ResponseHandler
import inspect


class TestResponseHandler:
    @pytest.mark.parametrize(
        "cors, expected_headers",
        [
            (
                True,
                b"Access-Control-Allow-Origin: *\r\n"
                b"Access-Control-Allow-Methods: GET, POST, OPTIONS\r\n"
                b"Access-Control-Allow-Headers: Content-Type\r\n",
            ),
            (False, b""),
        ],
    )
    def test_get_cors_headers(self, cors, expected_headers):
        response_handler = ResponseHandler(cors=cors)
        headers = response_handler._get_cors_headers()
        assert headers == expected_headers

    @pytest.mark.asyncio
    async def test_prepare_response_sse(self):
        async def async_generator():
            yield b"data: test\n\n"

        response_handler = ResponseHandler(cors=True)
        content_type = RouterBase.CONTENT_TYPE_SSE
        status_code = 200
        response_data = async_generator()

        response, generator = response_handler.prepare_response(
            content_type, status_code, response_data
        )

        assert response.startswith(b"HTTP/1.1 200 OK\r\n")
        assert b"Content-Type: text/event-stream\r\n" in response
        assert b"Access-Control-Allow-Origin: *\r\n" in response
        assert inspect.isasyncgen(generator)

    @pytest.mark.asyncio
    async def test_prepare_response_with_string(self):
        response_handler = ResponseHandler(cors=False)
        content_type = "text/plain"
        status_code = 200
        response_data = "Hello, world!"

        response = response_handler.prepare_response(
            content_type, status_code, response_data
        )

        assert response.startswith(b"HTTP/1.1 200 OK\r\n")
        assert b"Content-Type: text/plain\r\n" in response
        assert b"Content-Length: 13\r\n" in response
        assert b"Hello, world!" in response

    def test_prepare_options_response(self):
        response_handler = ResponseHandler(cors=True)
        response = response_handler.prepare_options_response()

        assert response.startswith(b"HTTP/1.1 204 No Content\r\n")
        assert b"Access-Control-Allow-Origin: *\r\n" in response
        assert b"Access-Control-Allow-Methods: GET, POST, OPTIONS\r\n" in response
        assert b"Access-Control-Allow-Headers: Content-Type\r\n" in response
        assert b"Content-Length: 0\r\n" in response
        assert b"Connection: close\r\n" in response

    def test_prepare_error_response_with_dict(self):
        response_handler = ResponseHandler(cors=False)
        status_code = 404
        response_data = {"error": "Not Found"}

        response = response_handler._prepare_error_response(status_code, response_data)

        assert response.startswith(b"HTTP/1.1 404 \r\n")
        assert b"Content-Type: application/json\r\n" in response
        assert b"Content-Length: 22\r\n" in response
        assert b'{"error": "Not Found"}' in response

    def test_prepare_normal_response_with_html_string(self):
        response_handler = ResponseHandler(cors=False)
        content_type = "text/html"
        status_code = 200
        response_data = "<html><body>Hello, world!</body></html>"

        response = response_handler._prepare_normal_response(
            content_type, status_code, response_data
        )

        assert response.startswith(b"HTTP/1.1 200 OK\r\n")
        assert b"Content-Type: text/html\r\n" in response
        assert b"Content-Length: 39\r\n" in response
        assert b"<html><body>Hello, world!</body></html>" in response

    def test_prepare_normal_response_with_json_dict(self):
        response_handler = ResponseHandler(cors=False)
        content_type = "application/json"
        status_code = 200
        response_data = {"message": "Hello, world!"}
        response_data_length = len(str(response_data).encode("utf-8"))

        response = response_handler._prepare_normal_response(
            content_type, status_code, response_data
        )

        assert response.startswith(b"HTTP/1.1 200 OK\r\n")
        assert b"Content-Type: application/json\r\n" in response
        assert f"Content-Length: {response_data_length}\r\n".encode("utf-8") in response
        assert b'{"message": "Hello, world!"}' in response



import pytest
from jinja2 import Environment
from bunnyhopapi.templates import (
    create_template_env,
    render_jinja_template,
    serve_static_file,
)
import os
import tempfile


class TestTemplates:
    def test_create_template_env(self):
        env = create_template_env("example/templates/jinja")
        assert isinstance(env, Environment)

    @pytest.mark.asyncio
    async def test_render_jinja_template_success(self):
        env = create_template_env("example/templates/jinja")
        status, content = await render_jinja_template("index.html", env, title="Test")
        assert status == 200
        assert "Test Page" in content

    @pytest.mark.asyncio
    async def test_render_jinja_template_not_found(self):
        env_path = "/"
        template_name = "nonexistent.html"
        env = create_template_env(env_path)
        status, content = await render_jinja_template(template_name, env)
        assert status == 404
        assert "Template not found" == content

    @pytest.mark.asyncio
    async def test_render_jinja_template_error(self):
        env = create_template_env("example/templates/jinja")
        status, content = await render_jinja_template(None, env)
        assert status == 500
        assert (
            "Internal server error: 'NoneType' object has no attribute 'split'"
            == content
        )

    @pytest.mark.asyncio
    async def test_serve_static_file_success(self):
        with tempfile.NamedTemporaryFile(delete=False) as tmp_file:
            tmp_file.write(b"Test content")
            tmp_file_path = tmp_file.name

        try:
            status, content = await serve_static_file(tmp_file_path)
            assert status == 200
            assert content == "Test content"
        finally:
            os.remove(tmp_file_path)

    @pytest.mark.asyncio
    async def test_serve_static_file_not_found(self):
        status, content = await serve_static_file("nonexistent_file.txt")
        assert status == 404
        assert content == "File not found"

    @pytest.mark.asyncio
    async def test_serve_static_file_internal_error(self):
        status, content = await serve_static_file(None)
        assert status == 500
        assert content == "Internal server error"



import pytest
from bunnyhopapi.swagger import SwaggerGenerator, SWAGGER_JSON
from bunnyhopapi.models import PathParam, QueryParam
from pydantic import BaseModel


class ExampleModel(BaseModel):
    name: str
    age: int


@pytest.fixture
def reset_swagger_json():
    SWAGGER_JSON["paths"] = {}
    SWAGGER_JSON["components"]["schemas"] = {}
    yield
    SWAGGER_JSON["paths"] = {}
    SWAGGER_JSON["components"]["schemas"] = {}


class TestSwaggerGenerator:
    def test_generate_path_item(self, reset_swagger_json):
        methods = {
            "GET": {
                "handler": lambda: None,
            }
        }
        SwaggerGenerator.generate_path_item("/test", methods)
        assert "/test" in SWAGGER_JSON["paths"]
        assert "get" in SWAGGER_JSON["paths"]["/test"]

    def test_generate_path_item_without_leading_slash(self, reset_swagger_json):
        methods = {
            "GET": {
                "handler": lambda: None,
            }
        }
        SwaggerGenerator.generate_path_item("test", methods)
        assert "/test" in SWAGGER_JSON["paths"]
        assert "get" in SWAGGER_JSON["paths"]["/test"]

    def test_process_path_params(self, reset_swagger_json):
        def handler(param: PathParam[int]):
            pass

        methods = {
            "GET": {
                "handler": handler,
            }
        }
        SwaggerGenerator.generate_path_item("/test/<param>", methods)
        assert "/test/{param}" in SWAGGER_JSON["paths"]
        parameters = SWAGGER_JSON["paths"]["/test/{param}"]["get"]["parameters"]
        assert len(parameters) == 1
        assert parameters[0]["name"] == "param"
        assert parameters[0]["in"] == "path"
        assert parameters[0]["schema"]["type"] == "integer"

    def test_process_query_params(self, reset_swagger_json):
        def handler(param: QueryParam[str] = "default"):
            pass

        methods = {
            "GET": {
                "handler": handler,
            }
        }
        SwaggerGenerator.generate_path_item("/test", methods)
        parameters = SWAGGER_JSON["paths"]["/test"]["get"]["parameters"]
        assert len(parameters) == 1
        assert parameters[0]["name"] == "param"
        assert parameters[0]["in"] == "query"
        assert parameters[0]["schema"]["type"] == "string"
        assert not parameters[0]["required"]

    def test_process_body_params(self, reset_swagger_json):
        def handler(body: ExampleModel):  # Actualización del nombre de la clase
            pass

        methods = {
            "POST": {
                "handler": handler,
            }
        }
        SwaggerGenerator.generate_path_item("/test", methods)
        request_body = SWAGGER_JSON["paths"]["/test"]["post"]["requestBody"]
        schema = request_body["content"]["application/json"]["schema"]
        assert "ExampleModel" in SWAGGER_JSON["components"]["schemas"]
        assert schema.get("title") == "ExampleModel"
        assert schema.get("type") == "object"
        assert "properties" in schema
        assert "name" in schema["properties"]
        assert "age" in schema["properties"]

    def test_process_body_params_without_components(self, reset_swagger_json):
        del SWAGGER_JSON["components"]

        class ExampleModel(BaseModel):  # Renombrado de TestModel a ExampleModel
            name: str
            age: int

        def handler(body: ExampleModel):  # Actualización del nombre de la clase
            pass

        methods = {
            "POST": {
                "handler": handler,
            }
        }

        SwaggerGenerator.generate_path_item("/test", methods)

        assert "components" in SWAGGER_JSON
        assert "schemas" in SWAGGER_JSON["components"]
        assert "ExampleModel" in SWAGGER_JSON["components"]["schemas"]

    def test_process_response_types(self, reset_swagger_json):
        def handler() -> {200: ExampleModel}:  # Actualización del nombre de la clase
            pass

        methods = {
            "GET": {
                "handler": handler,
            }
        }
        SwaggerGenerator.generate_path_item("/test", methods)
        responses = SWAGGER_JSON["paths"]["/test"]["get"]["responses"]
        assert 200 in responses
        assert (
            responses[200]["content"]["application/json"]["schema"]["$ref"]
            == "#/components/schemas/ExampleModel"
        )
        assert "ExampleModel" in SWAGGER_JSON["components"]["schemas"]

    def test_generate_path_item_with_non_callable_handler(self, reset_swagger_json):
        methods = {
            "GET": {
                "handler": "not_a_function",
            }
        }
        with pytest.raises(TypeError, match="not_a_function is not a callable object."):
            SwaggerGenerator.generate_path_item("/test", methods)

    def test_process_response_types_with_none_model(self, reset_swagger_json):
        def handler() -> {204: None}:
            pass

        methods = {
            "GET": {
                "handler": handler,
            }
        }
        SwaggerGenerator.generate_path_item("/test", methods)
        responses = SWAGGER_JSON["paths"]["/test"]["get"]["responses"]
        assert 204 in responses
        assert responses[204]["description"] == "Response with status 204"

    def test_process_response_types_with_base_model(self, reset_swagger_json):
        class ExampleModel(BaseModel):  # Renombrado de TestModel a ExampleModel
            name: str
            age: int

        def handler() -> ExampleModel:  # Actualización del nombre de la clase
            pass

        methods = {
            "GET": {
                "handler": handler,
            }
        }
        SwaggerGenerator.generate_path_item("/test", methods)
        responses = SWAGGER_JSON["paths"]["/test"]["get"]["responses"]
        assert 200 in responses
        assert (
            responses[200]["content"]["application/json"]["schema"]["$ref"]
            == "#/components/schemas/ExampleModel"
        )
        assert "ExampleModel" in SWAGGER_JSON["components"]["schemas"]

    def test_add_response_model_without_components(self, reset_swagger_json):
        del SWAGGER_JSON["components"]

        class ExampleModel(BaseModel):  # Renombrado de TestModel a ExampleModel
            name: str
            age: int

        response = SwaggerGenerator._add_response_model(
            200, ExampleModel
        )  # Actualización del nombre de la clase

        assert "components" in SWAGGER_JSON
        assert "schemas" in SWAGGER_JSON["components"]
        assert "ExampleModel" in SWAGGER_JSON["components"]["schemas"]

        assert 200 in response
        assert response[200]["description"] == "Response with status 200"
        assert (
            response[200]["content"]["application/json"]["schema"]["$ref"]
            == "#/components/schemas/ExampleModel"
        )

    def test_get_swagger_ui_html(self):
        html = SwaggerGenerator.get_swagger_ui_html()

        assert "<!DOCTYPE html>" in html
        assert "<html>" in html
        assert "<head>" in html
        assert "<body>" in html
        assert "</html>" in html

        assert '<div id="swagger-ui"></div>' in html

        assert "swagger-ui.css" in html
        assert "swagger-ui-bundle.js" in html



import asyncio
import hashlib
import base64
import struct
import uuid
from typing import Callable
import inspect
from . import logger


class WebSocketHandler:
    def __init__(self, websocket_handlers: dict[str, Callable]):
        self.websocket_handlers = websocket_handlers

    async def _read_websocket_frame(self, reader):
        header = await reader.read(2)
        if len(header) < 2:
            raise ConnectionResetError("Incomplete frame header")

        opcode = header[0] & 0x0F
        if opcode == 0x8:  # Close frame
            return opcode, None

        masked = (header[1] & 0x80) != 0
        payload_len = header[1] & 0x7F

        if payload_len == 126:
            payload_len = struct.unpack(">H", await reader.read(2))[0]
        elif payload_len == 127:
            payload_len = struct.unpack(">Q", await reader.read(8))[0]

        mask = await reader.read(4) if masked else None
        data = await reader.read(payload_len)

        if masked:
            decoded = bytearray(data)
            for i in range(len(decoded)):
                decoded[i] ^= mask[i % 4]
            data = decoded

        return opcode, data.decode("utf-8")

    async def _write_websocket_frame(self, writer, message):
        message_bytes = message.encode("utf-8")
        frame = bytearray()
        frame.append(0x81)
        frame.append(len(message_bytes))
        frame.extend(message_bytes)
        writer.write(frame)
        await writer.drain()

    async def handle_websocket(self, reader, writer, path, headers):
        if path not in self.websocket_handlers:
            logger.info(f"No WebSocket handler found for {path}")
            return

        handler_info = self.websocket_handlers[path]
        handler = handler_info["handler"]
        middleware = handler_info.get("middleware")
        connection = handler_info.get("connection")
        disconnect = handler_info.get("disconnect")

        key = headers.get("Sec-WebSocket-Key")
        if not key:
            logger.info("No Sec-WebSocket-Key found in headers")
            return

        if connection:
            is_auth = await connection(headers=headers)
            if not is_auth:
                writer.close()
                await writer.wait_closed()
                return

        accept_key = base64.b64encode(
            hashlib.sha1(
                key.encode() + b"258EAFA5-E914-47DA-95CA-C5AB0DC85B11"
            ).digest()
        ).decode()

        response = (
            "HTTP/1.1 101 Switching Protocols\r\n"
            "Upgrade: websocket\r\n"
            "Connection: Upgrade\r\n"
            f"Sec-WebSocket-Accept: {accept_key}\r\n\r\n"
        ).encode()

        writer.write(response)
        await writer.drain()

        connection_id = str(uuid.uuid4())

        async def process_message(message):
            if middleware:
                result = middleware(
                    connection_id=connection_id, message=message, headers=headers
                )
            else:
                result = handler(
                    connection_id=connection_id, message=message, headers=headers
                )

            if inspect.isasyncgen(result):
                async for response in result:
                    yield response
            else:
                yield await result

        try:
            while True:
                opcode, message = await self._read_websocket_frame(reader)
                if opcode == 0x8:  # Close frame
                    break

                async for response in process_message(message):
                    await self._write_websocket_frame(writer, response)

        except (ConnectionResetError, asyncio.IncompleteReadError) as e:
            logger.warning(f"WebSocket connection error: {e}")
            logger.info("WebSocket connection closed abruptly")
        finally:
            writer.close()
            await writer.wait_closed()

            if disconnect:
                logger.info("Executing disconnect middleware")
                await disconnect(connection_id=connection_id, headers=headers)
                logger.info("Disconnect middleware executed successfully")



import inspect
from typing import AsyncGenerator, TypeVar, Generic, Type, Coroutine, get_origin
import re
from functools import partial
from dataclasses import dataclass, field

T = TypeVar("T")


class PathParam(Generic[T]):
    def __init__(self, param_type: Type[T]):
        self.param_type = param_type

    def validate(self, value: str) -> T:
        try:
            return self.param_type(value)
        except (ValueError, TypeError):
            raise ValueError(f"Invalid value for type {self.param_type}: {value}")


class QueryParam(Generic[T]):
    def __init__(self, param_type: Type[T], required: bool = False):
        self.param_type = param_type
        self.required = required

    def validate(self, value: str) -> T:
        try:
            return self.param_type(value)
        except (ValueError, TypeError):
            raise ValueError(f"Invalid value for type {self.param_type}: {value}")


class BaseEndpoint:
    middleware = None

    @staticmethod
    def MIDDLEWARE():
        def decorator(func):
            setattr(func, "__class__middleware__", True)

            return func

        return decorator

    @staticmethod
    def __generic_endpoint_method(
        method_name: str,
        middleware=None,
        content_type=None,
    ):
        def decorator(func):
            setattr(func, "__http_method__", method_name)
            setattr(func, "__middleware__", middleware)
            setattr(func, "__content_type__", content_type)

            return func

        return decorator

    @staticmethod
    def GET(middleware=None, content_type=None):
        return BaseEndpoint.__generic_endpoint_method(
            "GET", middleware=middleware, content_type=content_type
        )

    @staticmethod
    def POST(middleware=None, content_type=None):
        return BaseEndpoint.__generic_endpoint_method(
            "POST", middleware=middleware, content_type=content_type
        )

    @staticmethod
    def PUT(middleware=None, content_type=None):
        return BaseEndpoint.__generic_endpoint_method(
            "PUT", middleware=middleware, content_type=content_type
        )

    @staticmethod
    def DELETE(middleware=None, content_type=None):
        return BaseEndpoint.__generic_endpoint_method(
            "DELETE", middleware=middleware, content_type=content_type
        )

    @staticmethod
    def PATCH(middleware=None, content_type=None):
        return BaseEndpoint.__generic_endpoint_method(
            "PATCH", middleware=middleware, content_type=content_type
        )


class Endpoint(BaseEndpoint):
    path: str = ""
    middleware: Coroutine = None

    def get_routes(self):
        routes = {}
        for method_name in dir(self):
            method = getattr(self, method_name)
            if callable(method) and getattr(method, "__class__middleware__", None):
                self.middleware = method
                break

        for method_name in dir(self):
            end_point = getattr(self, method_name)
            if callable(end_point):
                http_method = getattr(end_point, "__http_method__", None)
                if http_method:
                    annotations = getattr(end_point, "__annotations__", {})
                    if annotations:
                        param_names = "/".join(
                            f"<{name}>"
                            for name, value in annotations.items()
                            if get_origin(value) is PathParam
                        )

                        if not param_names:
                            route_path = self.path
                        else:
                            route_path = f"{self.path}/{param_names}".replace("//", "/")
                    else:
                        route_path = self.path

                    method_middleware = getattr(end_point, "__middleware__", None)

                    final_middleware = None

                    if self.middleware and method_middleware:
                        final_middleware = partial(
                            self.middleware,
                            endpoint=partial(method_middleware, endpoint=end_point),
                        )
                    elif self.middleware:
                        final_middleware = partial(self.middleware, endpoint=end_point)
                    elif method_middleware:
                        final_middleware = partial(
                            method_middleware, endpoint=end_point
                        )

                    if route_path not in routes:
                        routes[route_path] = {}

                    routes[route_path][http_method] = {
                        "handler": end_point,
                        "content_type": getattr(end_point, "__content_type__")
                        or RouterBase.CONTENT_TYPE_JSON,
                        "middleware": final_middleware,
                    }
        return routes

    def get_ws_routes(self):
        route_path = self.path
        routes = {}
        routes[route_path] = {}
        start_end_methods = ["connection", "disconnect"]
        for method_name in dir(self):
            if method_name == "ws":
                method = getattr(self, method_name)

                final_middleware = None

                if self.middleware:
                    final_middleware = partial(self.middleware, endpoint=method)

                if callable(method):
                    routes[route_path].update(
                        {
                            "handler": method,
                            "middleware": final_middleware,
                        }
                    )
            elif method_name in start_end_methods:
                method = getattr(self, method_name)
                if callable(method):
                    routes[route_path].update(
                        {
                            method_name: method,
                        }
                    )

        return routes


@dataclass(frozen=True)
class ServerConfig:
    port: int = 8000
    host: str = "0.0.0.0"
    cors: bool = False


@dataclass(frozen=True)
class RouterBase:
    routes: dict = field(default_factory=dict)
    routes_with_params: dict[str, re.Pattern] = field(default_factory=dict)
    middleware: Coroutine = field(default=None)
    websocket_handlers: dict[str, Coroutine] = field(default_factory=dict)
    prefix: str = field(default_factory=str)

    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"

    CONTENT_TYPE_JSON = "application/json"
    CONTENT_TYPE_HTML = "text/html"
    CONTENT_TYPE_SSE = "text/event-stream"

    def include_endpoint_class(self, endpoint_class: Type[Endpoint]):
        endpoint = endpoint_class()
        for path, context in endpoint.get_routes().items():
            for method, handler in context.items():
                self.add_route(
                    path=path,
                    method=method,
                    handler=handler.get("handler"),
                    middleware=handler.get("middleware"),
                    content_type=handler.get("content_type"),
                )

        for path, context in endpoint.get_ws_routes().items():
            self.add_websocket_route(
                path=path,
                handler=context.get("handler"),
                middleware=context.get("middleware"),
                connection=context.get("connection"),
                disconnect=context.get("disconnect"),
            )

    def include_router(self, router):
        for path, context in router.routes.items():
            full_path = f"/{self.prefix.lstrip('/')}/{path.lstrip('/')}".replace(
                "//", "/"
            )

            if full_path not in self.routes:
                self.routes[full_path] = {}
                if "<" in path:
                    self.routes_with_params[full_path] = self._compile_route_pattern(
                        full_path
                    )

            for method, content in context.items():
                handler = content.get("handler")
                existing_middleware = content.get("middleware")

                middleware = None
                if self.middleware and existing_middleware:
                    middleware = partial(
                        self.middleware,
                        endpoint=existing_middleware,
                    )
                elif self.middleware:
                    middleware = partial(self.middleware, endpoint=handler)
                elif existing_middleware:
                    middleware = existing_middleware

                self.routes[full_path][method] = {
                    "handler": handler,
                    "content_type": content.get("content_type"),
                    "middleware": middleware,
                }

        self.routes_with_params.update(router.routes_with_params)
        self.websocket_handlers.update(router.websocket_handlers)

    def add_route(
        self,
        path: str,
        method: str,
        handler: Coroutine,
        content_type=CONTENT_TYPE_JSON,
        middleware: Coroutine = None,
    ):
        path = path.lstrip("/")
        full_path = f"/{self.prefix.lstrip('/')}/{path}".replace("//", "/")

        if full_path not in self.routes:
            self.routes[full_path] = {}
            if "<" in path:
                self.routes_with_params[full_path] = self._compile_route_pattern(
                    full_path
                )

        final_middleware = None
        if middleware and self.middleware:
            final_middleware = partial(
                self.middleware,
                endpoint=partial(middleware, endpoint=handler),
            )
        elif middleware:
            final_middleware = partial(middleware, endpoint=handler)
        elif self.middleware:
            final_middleware = partial(self.middleware, endpoint=handler)

        self.routes[full_path][method] = {
            "handler": handler,
            "middleware": final_middleware,
            "content_type": content_type,
        }

    def _compile_route_pattern(self, path: str) -> re.Pattern:
        param_pattern = re.compile(r"<(\w+)>")
        regex_pattern = re.sub(param_pattern, r"(?P<\1>[^/]+)", path)
        return re.compile(regex_pattern + r"/?$")

    def add_websocket_route(
        self,
        path,
        handler,
        middleware: AsyncGenerator = None,
        connection: Coroutine = None,
        disconnect: Coroutine = None,
    ):
        class_middleware = (
            self.middleware if inspect.isasyncgenfunction(self.middleware) else None
        )
        method_middleware = (
            middleware if inspect.isasyncgenfunction(middleware) else None
        )

        final_middleware = None

        if class_middleware and method_middleware:
            final_middleware = partial(
                class_middleware,
                endpoint=partial(method_middleware, endpoint=handler),
            )
        elif class_middleware:
            final_middleware = partial(class_middleware, endpoint=handler)

        elif method_middleware:
            final_middleware = partial(method_middleware, endpoint=handler)

        self.websocket_handlers[path] = {
            "handler": handler,
            "middleware": final_middleware,
            "connection": connection,
            "disconnect": disconnect,
        }



from jinja2 import Environment, FileSystemLoader, TemplateNotFound
from . import logger


def create_template_env(path: str):
    return Environment(loader=FileSystemLoader(path))


async def render_jinja_template(
    template_name: str, template_env: Environment, **context
):
    try:
        template = template_env.get_template(template_name)
        return 200, template.render(**context)
    except TemplateNotFound:
        return 404, "Template not found"
    except Exception as e:
        logger.info(f"Error rendering template {template_name}: {e}")
        logger.info(f"error type: {type(e)}")
        return 500, f"Internal server error: {e}"


async def serve_static_file(file_path: str, *args, **kwargs):
    try:
        with open(file_path, "r", encoding="utf-8") as static_file:
            return 200, static_file.read()
    except FileNotFoundError:
        logger.error(f"file not found: {file_path}")
        return 404, "File not found"
    except Exception as e:
        logger.error(f"Error reading file {file_path}: {e}")
        return 500, "Internal server error"



import asyncio
import os
import mimetypes
import signal
from multiprocessing import Process
from .templates import serve_static_file
from .models import ServerConfig, RouterBase
from .swagger import SwaggerGenerator, SWAGGER_JSON
from .client_handler import ClientHandler
from . import logger
import uvloop
from dataclasses import dataclass


@dataclass(frozen=True, slots=True)
class Router(RouterBase):
    pass


@dataclass(frozen=True, slots=True)
class Server(ServerConfig, RouterBase):
    async def generate_swagger_json(self, *args, **kwargs):
        if not SWAGGER_JSON["paths"]:
            for path, methods in self.routes.items():
                if path in {"/docs", "/swagger.json"}:
                    continue
                SwaggerGenerator.generate_path_item(path, methods)
        return 200, SWAGGER_JSON

    async def swagger_ui_handler(self, *args, **kwargs):
        return 200, SwaggerGenerator.get_swagger_ui_html()

    def add_swagger(self):
        self.add_route(
            path="/swagger.json",
            method="GET",
            handler=self.generate_swagger_json,
            content_type="application/json",
        )
        self.add_route(
            path="/docs",
            method="GET",
            handler=self.swagger_ui_handler,
            content_type="text/html",
        )

    def include_static_folder(self, folder_path: str, route_prefix: str = "/static"):
        folder_path = os.path.abspath(folder_path)

        if not os.path.exists(folder_path):
            logger.error(f"Static folder does not exist: {folder_path}")
            return

        if not os.path.isdir(folder_path):
            logger.error(f"Provided path is not a directory: {folder_path}")
            return

        for root, _, files in os.walk(folder_path):
            logger.debug(f"Walking through: {root}")
            for file_name in files:
                file_path = os.path.join(root, file_name)
                relative_path = os.path.relpath(file_path, folder_path)
                route = f"{route_prefix}/{relative_path}".replace("\\", "/")
                logger.info(f"Serving static file: {file_path} -> {route}")
                content_type, _ = mimetypes.guess_type(file_path)
                logger.info(f"Content type for {file_path}: {content_type}")

                self.add_route(
                    path=route,
                    method="GET",
                    content_type=content_type,
                    handler=lambda headers, file_path=file_path: serve_static_file(
                        file_path
                    ),
                )

    async def _run(self):
        client_handler = ClientHandler(
            self.routes, self.routes_with_params, self.websocket_handlers, self.cors
        )

        server = await asyncio.start_server(
            client_handler.handle_client,
            self.host,
            self.port,
            reuse_address=True,
            reuse_port=True,
        )

        addr = server.sockets[0].getsockname()
        logger.info(f"HTTP server listening on {addr}")

        async with server:
            await server.serve_forever()

    def _start_worker(self):
        try:
            loop = asyncio.get_event_loop()
            if loop.is_running():
                loop.create_task(self._run())
            else:
                asyncio.run(self._run())
        except KeyboardInterrupt:
            logger.info("Worker received stop signal")

    def run(self, workers: int = os.cpu_count() or 1):
        self.add_swagger()

        uvloop.install()
        processes = []

        try:
            for _ in range(workers):
                p = Process(target=self._start_worker)
                p.start()
                processes.append(p)

            for p in processes:
                p.join()
        except KeyboardInterrupt:
            logger.info("Server stopped by user")
        finally:
            for p in processes:
                p.terminate()
            logger.info("Server stopped")


def handle_exit(signal, frame):
    logger.info("Shutting down server...")
    os._exit(0)


signal.signal(signal.SIGINT, handle_exit)
signal.signal(signal.SIGTERM, handle_exit)



from setuptools import setup, find_packages

setup(
    name="bunnyhopapi",
    version="1.3.4",
    description="A simple, easy and fast http server for python",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Ricardo Robles Fernández",
    author_email="ricardo.r.f@hotmail.com",
    url="https://github.com/rroblf01/bunnyhopapi",
    packages=find_packages(),
    python_requires=">=3.10",
    install_requires=[
        "pydantic==2.11.0",
    ],
    extras_require={
        "lint": ["ruff==0.11.2"],
    },
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    entry_points={
        "console_scripts": [
            "bunnyhopapi-example=example.main:main",
        ],
    },
)



import json
from typing import Union, AsyncGenerator, Tuple
import inspect
from . import logger
from .models import RouterBase


class ResponseHandler:
    def __init__(self, cors: bool = False):
        self.cors = cors

    def _get_cors_headers(self):
        if not self.cors:
            return b""

        return (
            b"Access-Control-Allow-Origin: *\r\n"
            b"Access-Control-Allow-Methods: GET, POST, OPTIONS\r\n"
            b"Access-Control-Allow-Headers: Content-Type\r\n"
        )

    def prepare_response(
        self,
        content_type: str,
        status_code: int,
        response_data: Union[dict, str, bytes, AsyncGenerator],
    ) -> Union[Tuple[bytes, AsyncGenerator], bytes]:
        if content_type == RouterBase.CONTENT_TYPE_SSE and inspect.isasyncgen(
            response_data
        ):
            return self._prepare_sse_response(response_data)

        if isinstance(response_data, dict):
            response_data = json.dumps(response_data).encode("utf-8")
        elif isinstance(response_data, str):
            response_data = response_data.encode("utf-8")

        if status_code >= 400:
            return self._prepare_error_response(status_code, response_data)

        return self._prepare_normal_response(content_type, status_code, response_data)

    def _prepare_sse_response(self, generator: AsyncGenerator):
        response = (
            (
                b"HTTP/1.1 200 OK\r\n"
                b"Content-Type: text/event-stream\r\n"
                b"Cache-Control: no-cache\r\n"
                b"Connection: keep-alive\r\n"
            )
            + self._get_cors_headers()
            + b"\r\n"
        )

        return response, generator

    def _prepare_error_response(
        self, status_code: int, response_data: Union[dict, str]
    ):
        content_type = "application/json"
        if isinstance(response_data, dict):
            response_data = json.dumps(response_data).encode("utf-8")

        return self._build_response(content_type, status_code, response_data)

    def _prepare_normal_response(
        self,
        content_type: str,
        status_code: int,
        response_data: Union[dict, str, bytes],
    ):
        if content_type == "text/html" and isinstance(response_data, str):
            response_data = response_data.encode("utf-8")
        elif content_type == "application/json" and isinstance(response_data, dict):
            response_data = json.dumps(response_data).encode("utf-8")

        return self._build_response(content_type, status_code, response_data)

    def _build_response(
        self, content_type: str, status_code: int, response_data: bytes
    ):
        content_length = len(response_data) if response_data else 0
        status_text = "OK" if status_code == 200 else ""

        return (
            (
                b"HTTP/1.1 "
                + str(status_code).encode("utf-8")
                + b" "
                + status_text.encode("utf-8")
                + b"\r\n"
                b"Content-Type: " + content_type.encode("utf-8") + b"\r\n"
                b"Content-Length: " + str(content_length).encode("utf-8") + b"\r\n"
            )
            + self._get_cors_headers()
            + b"Connection: close\r\n\r\n"
            + (response_data if response_data else b"")
        )

    def prepare_options_response(self):
        return (
            b"HTTP/1.1 204 No Content\r\n"
            b"Access-Control-Allow-Origin: *\r\n"
            b"Access-Control-Allow-Methods: GET, POST, OPTIONS\r\n"
            b"Access-Control-Allow-Headers: Content-Type\r\n"
            b"Content-Length: 0\r\n"
            b"Connection: close\r\n"
            b"\r\n"
        )



import json
import inspect
from . import logger
from typing import get_type_hints, get_origin
import asyncio
from pydantic import BaseModel, ValidationError
from bunnyhopapi.models import PathParam, QueryParam
from .models import RouterBase
from dataclasses import dataclass, field


@dataclass
class RouteHandler:
    routes: dict = field(default_factory=dict)
    routes_with_params: dict = field(default_factory=dict)

    def _extract_params(self, path: str, route_path: str):
        if route_path not in self.routes_with_params:
            return None

        compiled_pattern = self.routes_with_params[route_path]
        match = compiled_pattern.match(path)
        if not match:
            return None

        return match.groupdict()

    async def execute_handler(
        self,
        path: str,
        method: str,
        body: str | None = None,
        headers: dict | None = None,
        query_params: dict | None = None,
    ):
        route_info = self._find_route(path, method)
        if not route_info:
            return await self._route_not_found_response(path, method)

        handler_info, params = route_info
        handler_info["params"] = params
        return await self._process_handler(
            handler_info, path, method, body, headers, query_params
        )

    def _find_route(self, path: str, method: str):
        if path in self.routes and method in self.routes[path]:
            return self.routes[path][method], {}

        for route_path, methods in self.routes.items():
            if method not in methods:
                continue

            params = self._extract_params(path, route_path)
            if params is not None:
                return methods[method], params

        return None

    async def _route_not_found_response(self, path: str, method: str):
        error_msg = {"error": f"Route {method} {path} not found"}
        return "application/json", 404, error_msg

    async def _process_handler(
        self,
        route_info: dict,
        path: str,
        method: str,
        body: str = None,
        headers: dict = None,
        query_params: dict = None,
    ):
        handler = route_info["handler"]
        content_type = route_info["content_type"]
        middleware = route_info.get("middleware")
        route_params = route_info.get("params", {})

        type_hints = get_type_hints(handler)

        try:
            validated_params = self._validate_params(
                route_params, type_hints, query_params
            )
        except ValueError as e:
            logger.error(f"Validation error for {method} {path}: {str(e)}")
            return {
                "content_type": content_type,
                "status_code": 422,
                "response_data": {"error": str(e)},
            }

        if body:
            body_validation = self._validate_body(body, type_hints)
            if body_validation.get("error"):
                return {
                    "content_type": content_type,
                    "status_code": body_validation["status_code"],
                    "response_data": body_validation["error"],
                }
            validated_params.update(body_validation["validated_params"])

        try:
            result = (middleware or handler)(headers=headers, **validated_params)

            if inspect.isasyncgen(result):
                return {
                    "content_type": RouterBase.CONTENT_TYPE_SSE,
                    "status_code": 200,
                    "response_data": result,
                }

            response = await result if asyncio.iscoroutine(result) else result

            if not isinstance(response, tuple) or len(response) != 2:
                raise ValueError(
                    "Handler must return a tuple of (status_code, response_data)"
                )

            status_code, response_data = response

            return_type = type_hints.get("return", {})
            if isinstance(return_type, dict) and (
                response_model := return_type.get(status_code)
            ):
                try:
                    response_data = response_model.model_validate(
                        response_data
                    ).model_dump()
                except ValidationError as e:
                    logger.error(f"Validation error: {e}", exc_info=True)
                    return {
                        "content_type": content_type,
                        "status_code": 422,
                        "response_data": {
                            "error": "Validation error",
                            "details": e.errors(),
                        },
                    }

            return {
                "content_type": content_type,
                "status_code": status_code,
                "response_data": response_data,
            }

        except Exception as e:
            logger.error(f"Handler error for {method} {path}: {str(e)}", exc_info=True)
            return {
                "content_type": content_type,
                "status_code": 500,
                "response_data": {"error": "Internal server error", "message": str(e)},
            }

    def _validate_params(self, params: dict, type_hints: dict, query_params: dict = {}):
        validated_params = {}
        combined_params = {**params, **(query_params or {})}

        for param_name, param_value in combined_params.items():
            if param_name not in type_hints:
                continue

            param_type = type_hints[param_name]
            origin = get_origin(param_type)

            try:
                if origin is PathParam or origin is QueryParam:
                    param_class = param_type.__args__[0]
                    default_value = (
                        param_type.__args__[1] if len(param_type.__args__) > 1 else None
                    )
                    validated_params[param_name] = (
                        param_class(param_value)
                        if param_value is not None
                        else default_value
                    )
                else:
                    validated_params[param_name] = param_value
            except (ValueError, TypeError):
                raise ValueError(
                    f"Invalid value for parameter '{param_name}': '{param_value}' is not a valid {param_class.__name__}."
                )

        return validated_params

    def _validate_body(self, body: str, type_hints: dict):
        try:
            body_data = json.loads(body)
        except json.JSONDecodeError:
            return {
                "status_code": 400,
                "error": {"error": "Invalid JSON format in request body"},
                "validated_params": {},
            }

        validated_params = {}
        for param_name, param_type in type_hints.items():
            if (
                inspect.isclass(param_type)
                and issubclass(param_type, BaseModel)
                and param_name not in validated_params
            ):
                try:
                    validated_params[param_name] = param_type(**body_data)
                except ValueError as e:
                    errors = e.errors()
                    error_details = [
                        {
                            "field": "->".join(str(loc) for loc in error["loc"]),
                            "message": error["msg"],
                            "type": error["type"],
                        }
                        for error in errors
                    ]
                    return {
                        "status_code": 422,
                        "error": {
                            "error": "Validation error",
                            "details": error_details,
                        },
                        "validated_params": {},
                    }

        return {"status_code": 200, "error": None, "validated_params": validated_params}



import pytest
from unittest.mock import AsyncMock, Mock, patch
from bunnyhopapi.client_handler import ClientHandler
from bunnyhopapi.models import RouterBase


class TestClientHandler:
    @pytest.fixture
    def client_handler(self):
        routes = {}
        routes_with_params = {}
        websocket_handlers = {}
        return ClientHandler(routes, routes_with_params, websocket_handlers, cors=True)

    @pytest.mark.asyncio
    async def test_handle_client_invalid_request(self, client_handler):
        reader = AsyncMock()
        writer = AsyncMock()
        reader.read = AsyncMock(return_value=b"")
        writer.write = Mock()
        writer.close = Mock()

        await client_handler.handle_client(reader, writer)

        writer.write.assert_called_once()
        assert b"400" in writer.write.call_args[0][0]

    @pytest.mark.asyncio
    async def test_handle_client_websocket_request(self, client_handler):
        reader = AsyncMock()
        writer = AsyncMock()
        reader.read = AsyncMock(
            return_value=b"GET /ws HTTP/1.1\r\nConnection: Upgrade\r\nUpgrade: websocket\r\n\r\n"
        )
        writer.write = Mock()

        with patch.object(
            client_handler.websocket_handler, "handle_websocket", new=AsyncMock()
        ) as mock_ws_handler:
            await client_handler.handle_client(reader, writer)

        mock_ws_handler.assert_called_once()

    @pytest.mark.asyncio
    async def test_send_response(self, client_handler):
        writer = AsyncMock()
        writer.write = Mock()
        writer.drain = AsyncMock()
        writer.close = Mock()
        writer.wait_closed = AsyncMock()

        response = {
            "content_type": "application/json",
            "status_code": 200,
            "response_data": {"message": "Success"},
        }

        with patch.object(
            client_handler.response_handler,
            "prepare_response",
            return_value=b"HTTP/1.1 200 OK\r\n\r\n",
        ) as mock_prepare_response:
            await client_handler._send_response(writer, response)

        mock_prepare_response.assert_called_once_with(
            "application/json", 200, {"message": "Success"}
        )
        writer.write.assert_called_once_with(b"HTTP/1.1 200 OK\r\n\r\n")
        writer.drain.assert_called_once()
        writer.close.assert_called_once()
        writer.wait_closed.assert_called_once()

    @pytest.mark.asyncio
    async def test_send_response_sse(self, client_handler):
        writer = AsyncMock()
        writer.write = Mock()
        writer.drain = AsyncMock()
        writer.close = Mock()
        writer.wait_closed = AsyncMock()

        async def async_gen():
            yield "data: message1\n\n"
            yield "data: message2\n\n"

        response = {
            "content_type": RouterBase.CONTENT_TYPE_SSE,
            "status_code": 200,
            "response_data": async_gen(),
        }

        with patch.object(
            client_handler.response_handler,
            "prepare_response",
            return_value=(
                b"HTTP/1.1 200 OK\r\nContent-Type: text/event-stream\r\n\r\n",
                async_gen(),
            ),
        ) as mock_prepare_response:
            await client_handler._send_response(writer, response)

        mock_prepare_response.assert_called_once_with(
            RouterBase.CONTENT_TYPE_SSE, 200, response["response_data"]
        )
        writer.write.assert_any_call(
            b"HTTP/1.1 200 OK\r\nContent-Type: text/event-stream\r\n\r\n"
        )
        writer.write.assert_any_call(b"data: message1\n\n")
        writer.write.assert_any_call(b"data: message2\n\n")
        writer.drain.assert_called()
        writer.close.call_count == 2
        writer.wait_closed.call_count == 2

    @pytest.mark.asyncio
    async def test_handle_options(self, client_handler):
        writer = AsyncMock()
        writer.write = Mock()
        writer.drain = AsyncMock()
        writer.close = Mock()
        writer.wait_closed = AsyncMock()

        with patch.object(
            client_handler.response_handler,
            "prepare_options_response",
            return_value=b"HTTP/1.1 204 No Content\r\n\r\n",
        ) as mock_prepare_options_response:
            await client_handler._handle_options(writer)

        mock_prepare_options_response.assert_called_once()
        writer.write.assert_called_once_with(b"HTTP/1.1 204 No Content\r\n\r\n")
        writer.drain.assert_called_once()
        writer.close.assert_called_once()
        writer.wait_closed.assert_called_once()

    @pytest.mark.asyncio
    async def test_send_response_tuple(self, client_handler):
        writer = AsyncMock()
        writer.write = Mock()
        writer.drain = AsyncMock()
        writer.close = Mock()
        writer.wait_closed = AsyncMock()

        response = ("application/json", 201, {"message": "Created"})

        with patch.object(
            client_handler.response_handler,
            "prepare_response",
            return_value=b"HTTP/1.1 201 Created\r\n\r\n",
        ) as mock_prepare_response:
            await client_handler._send_response(writer, response)

        mock_prepare_response.assert_called_once_with(
            "application/json", 201, {"message": "Created"}
        )
        writer.write.assert_called_once_with(b"HTTP/1.1 201 Created\r\n\r\n")
        writer.drain.assert_called_once()
        writer.close.assert_called_once()
        writer.wait_closed.assert_called_once()

    @pytest.mark.asyncio
    async def test_send_response_else_case(self, client_handler):
        writer = AsyncMock()
        writer.write = Mock()
        writer.drain = AsyncMock()
        writer.close = Mock()
        writer.wait_closed = AsyncMock()

        response = "invalid_response_format"

        with patch.object(
            client_handler.response_handler,
            "prepare_response",
            return_value=b"HTTP/1.1 500 Internal Server Error\r\n\r\n",
        ) as mock_prepare_response:
            await client_handler._send_response(writer, response)

        mock_prepare_response.assert_called_once_with(
            "application/json",
            500,
            {
                "error": "Internal server error",
                "message": "Unknown response format",
            },
        )
        writer.write.assert_called_once_with(
            b"HTTP/1.1 500 Internal Server Error\r\n\r\n"
        )
        writer.drain.assert_called_once()
        writer.close.assert_called_once()
        writer.wait_closed.assert_called_once()

    @pytest.mark.asyncio
    async def test_send_response_exception_handling(self, client_handler):
        writer = AsyncMock()
        writer.write = Mock(side_effect=Exception("Mocked write exception"))
        writer.drain = AsyncMock()
        writer.close = Mock()
        writer.wait_closed = AsyncMock()

        response = {
            "content_type": "application/json",
            "status_code": 200,
            "response_data": {"message": "Success"},
        }

        with patch("bunnyhopapi.client_handler.logger.error") as mock_logger_error:
            await client_handler._send_response(writer, response)

        mock_logger_error.assert_called_once_with(
            "Error sending response: Mocked write exception"
        )
        writer.close.assert_called_once()
        writer.wait_closed.assert_called_once()

    @pytest.mark.asyncio
    async def test_read_request_exception_handling(self, client_handler):
        reader = AsyncMock()
        reader.read = AsyncMock(side_effect=Exception("Mocked read exception"))

        with patch("bunnyhopapi.client_handler.logger.error") as mock_logger_error:
            result = await client_handler._read_request(reader)

        mock_logger_error.assert_called_once_with(
            "Error reading from client: Mocked read exception"
        )
        assert result is None

    @pytest.mark.asyncio
    async def test_handle_client_request_data_none(self, client_handler):
        reader = AsyncMock()
        writer = AsyncMock()
        reader.read = AsyncMock(return_value=None)
        writer.write = Mock()
        writer.close = Mock()

        with (
            patch.object(
                client_handler, "_read_request", return_value=None
            ) as mock_read_request,
            patch.object(
                client_handler, "_send_error_response", new=AsyncMock()
            ) as mock_send_error_response,
        ):
            await client_handler.handle_client(reader, writer)

        mock_read_request.assert_called_once_with(reader)
        mock_send_error_response.assert_called_once_with(writer, 400, "Invalid request")

    @pytest.mark.asyncio
    async def test_handle_client_options_method(self, client_handler):
        reader = AsyncMock()
        writer = AsyncMock()
        reader.read = AsyncMock(return_value=b"OPTIONS / HTTP/1.1\r\n\r\n")
        writer.write = Mock()
        writer.close = Mock()

        with (
            patch.object(
                client_handler.request_parser,
                "parse_request",
                return_value=("OPTIONS", "/", {}, None, {}),
            ) as mock_parse_request,
            patch.object(
                client_handler, "_handle_options", new=AsyncMock()
            ) as mock_handle_options,
        ):
            await client_handler.handle_client(reader, writer)

        mock_parse_request.assert_called_once_with(b"OPTIONS / HTTP/1.1\r\n\r\n")
        mock_handle_options.assert_called_once_with(writer)

    @pytest.mark.asyncio
    async def test_handle_client_execute_handler(self, client_handler):
        reader = AsyncMock()
        writer = AsyncMock()
        reader.read = AsyncMock(return_value=b"GET /test HTTP/1.1\r\n\r\n")
        writer.write = Mock()
        writer.close = Mock()

        with (
            patch.object(
                client_handler.request_parser,
                "parse_request",
                return_value=("GET", "/test", {}, None, {}),
            ) as mock_parse_request,
            patch.object(
                client_handler.route_handler,
                "execute_handler",
                new=AsyncMock(
                    return_value={
                        "content_type": "application/json",
                        "status_code": 200,
                        "response_data": {"message": "OK"},
                    }
                ),
            ) as mock_execute_handler,
            patch.object(
                client_handler, "_send_response", new=AsyncMock()
            ) as mock_send_response,
        ):
            await client_handler.handle_client(reader, writer)

        mock_parse_request.assert_called_once_with(b"GET /test HTTP/1.1\r\n\r\n")
        mock_execute_handler.assert_called_once_with("/test", "GET", None, {}, {})
        mock_send_response.assert_called_once_with(
            writer,
            {
                "content_type": "application/json",
                "status_code": 200,
                "response_data": {"message": "OK"},
            },
        )



import pytest
from bunnyhopapi.models import PathParam, QueryParam, Endpoint, RouterBase
import re
import asyncio
from bunnyhopapi import logger


class TestPathParam:
    def test_validate_valid_value(self):
        param = PathParam(int)
        assert param.validate("123") == 123

    def test_validate_invalid_value(self):
        param = PathParam(int)
        with pytest.raises(
            ValueError, match="Invalid value for type <class 'int'>: abc"
        ):
            param.validate("abc")


class TestQueryParam:
    def test_validate_valid_value(self):
        param = QueryParam(str)
        assert param.validate("test") == "test"

    def test_validate_invalid_value(self):
        param = QueryParam(int)
        with pytest.raises(
            ValueError, match="Invalid value for type <class 'int'>: test"
        ):
            param.validate("test")


class TestRouter:
    def test_compile_route_pattern(self):
        router = RouterBase()
        pattern = router._compile_route_pattern("/test/<id>")
        assert isinstance(pattern, re.Pattern)
        assert pattern.match("/test/123")
        assert not pattern.match("/test/abc/extra")

    def test_add_route(self):
        router = RouterBase()
        handler = lambda: None
        router.add_route("/test", "GET", handler)
        assert "/test" in router.routes
        assert "GET" in router.routes["/test"]
        assert router.routes["/test"]["GET"]["handler"] == handler

    def test_add_route_with_path_params(self):
        router = RouterBase()
        handler = lambda: None
        router.add_route("/test/<id>", "GET", handler)
        assert "/test/<id>" in router.routes_with_params
        assert isinstance(router.routes_with_params["/test/<id>"], re.Pattern)

    def test_add_route_with_both_middlewares(self):
        async def class_middleware(endpoint, *args, **kwargs):
            return await endpoint(*args, **kwargs)

        async def method_middleware(endpoint, *args, **kwargs):
            return await endpoint(*args, **kwargs)

        async def handler(*args, **kwargs):
            return "handler executed"

        router = RouterBase(middleware=class_middleware)
        router.add_route("/test", "GET", handler, middleware=method_middleware)

        assert "/test" in router.routes
        assert "GET" in router.routes["/test"]
        assert callable(router.routes["/test"]["GET"]["middleware"])

    def test_add_route_with_method_middleware(self):
        async def method_middleware(endpoint, *args, **kwargs):
            return await endpoint(*args, **kwargs)

        async def handler(*args, **kwargs):
            return "handler executed"

        router = RouterBase()
        router.add_route("/test", "GET", handler, middleware=method_middleware)

        assert "/test" in router.routes
        assert "GET" in router.routes["/test"]
        assert callable(router.routes["/test"]["GET"]["middleware"])

    def test_add_route_with_class_middleware(self):
        async def class_middleware(endpoint, *args, **kwargs):
            return await endpoint(*args, **kwargs)

        async def handler(*args, **kwargs):
            return "handler executed"

        router = RouterBase(middleware=class_middleware)
        router.add_route("/test", "GET", handler)

        assert "/test" in router.routes
        assert "GET" in router.routes["/test"]
        assert callable(router.routes["/test"]["GET"]["middleware"])

    def test_add_websocket_route(self):
        router = RouterBase()
        handler = lambda: None
        router.add_websocket_route("/ws", handler)
        assert "/ws" in router.websocket_handlers
        assert router.websocket_handlers["/ws"]["handler"] == handler

    def test_add_websocket_route_with_class_and_method_middleware(self):
        async def class_middleware(endpoint, *args, **kwargs):
            return await endpoint(*args, **kwargs)

        async def method_middleware(endpoint, *args, **kwargs):
            yield await endpoint(*args, **kwargs)

        async def handler(*args, **kwargs):
            return "handler executed"

        router = RouterBase(middleware=class_middleware)
        router.add_websocket_route("/ws", handler, middleware=method_middleware)

        assert "/ws" in router.websocket_handlers
        assert router.websocket_handlers["/ws"]["middleware"] is not None

        combined_middleware = router.websocket_handlers["/ws"]["middleware"]
        assert callable(combined_middleware)

    def test_add_websocket_route_with_class_middleware(self):
        async def class_middleware(endpoint, *args, **kwargs):
            yield await endpoint(*args, **kwargs)

        async def handler(*args, **kwargs):
            return "handler executed"

        router = RouterBase(middleware=class_middleware)
        router.add_websocket_route("/ws", handler)

        assert "/ws" in router.websocket_handlers
        assert callable(router.websocket_handlers["/ws"]["middleware"])
        assert router.websocket_handlers["/ws"]["middleware"].func == class_middleware

    def test_add_websocket_route_with_method_middleware(self):
        async def method_middleware(endpoint, *args, **kwargs):
            yield await endpoint(*args, **kwargs)

        async def handler(*args, **kwargs):
            return "handler executed"

        router = RouterBase()
        router.add_websocket_route("/ws", handler, middleware=method_middleware)

        assert "/ws" in router.websocket_handlers
        assert callable(router.websocket_handlers["/ws"]["middleware"])
        assert router.websocket_handlers["/ws"]["middleware"].func == method_middleware

    def test_add_websocket_route_with_both_middlewares(self):
        async def class_middleware(endpoint, *args, **kwargs):
            yield await endpoint(*args, **kwargs)

        async def method_middleware(endpoint, *args, **kwargs):
            yield await endpoint(*args, **kwargs)

        async def handler(*args, **kwargs):
            return "handler executed"

        router = RouterBase(middleware=class_middleware)
        router.add_websocket_route("/ws", handler, middleware=method_middleware)

        assert "/ws" in router.websocket_handlers
        assert router.websocket_handlers["/ws"]["middleware"] is not None

        combined_middleware = router.websocket_handlers["/ws"]["middleware"]
        assert callable(combined_middleware)
        assert combined_middleware.func == class_middleware

    def test_include_endpoint_class_with_ws_endpoint(self):
        router = RouterBase()
        router.include_endpoint_class(WSEndpoint)

        assert "/ws/chat" in router.websocket_handlers
        ws_handler = router.websocket_handlers["/ws/chat"]

        assert "handler" in ws_handler
        assert callable(ws_handler["handler"])

        assert "connection" in ws_handler
        assert callable(ws_handler["connection"])

        assert "disconnect" in ws_handler
        assert callable(ws_handler["disconnect"])

        assert "middleware" in ws_handler
        assert callable(ws_handler["middleware"])

    def test_include_router(self):
        router1 = RouterBase()
        router2 = RouterBase()

        handler1 = lambda: None
        handler2 = lambda: None

        router1.add_route("/test1", "GET", handler1)
        router2.add_route("/test2", "POST", handler2)

        router1.include_router(router2)

        assert "/test1" in router1.routes
        assert "GET" in router1.routes["/test1"]
        assert router1.routes["/test1"]["GET"]["handler"] == handler1

        assert "/test2" in router1.routes
        assert "POST" in router1.routes["/test2"]
        assert router1.routes["/test2"]["POST"]["handler"] == handler2

    def test_include_router_with_prefix(self):
        router1 = RouterBase(prefix="/api")
        router2 = RouterBase()

        handler = lambda: None
        router2.add_route("/test", "GET", handler)

        router1.include_router(router2)

        assert "/api/test" in router1.routes
        assert "GET" in router1.routes["/api/test"]
        assert router1.routes["/api/test"]["GET"]["handler"] == handler

    def test_include_router_with_path_params(self):
        router1 = RouterBase()
        router2 = RouterBase()

        handler = lambda: None
        router2.add_route("/test/<id>", "GET", handler)

        router1.include_router(router2)

        assert "/test/<id>" in router1.routes_with_params
        assert isinstance(router1.routes_with_params["/test/<id>"], re.Pattern)

    def test_include_router_with_both_middlewares(self):
        async def class_middleware(endpoint, *args, **kwargs):
            return await endpoint(*args, **kwargs)

        async def method_middleware(endpoint, *args, **kwargs):
            return await endpoint(*args, **kwargs)

        async def handler(*args, **kwargs):
            return "handler executed"

        router1 = RouterBase(middleware=class_middleware)
        router2 = RouterBase()
        router2.add_route("/test", "GET", handler, middleware=method_middleware)

        router1.include_router(router2)

        assert "/test" in router1.routes
        assert "GET" in router1.routes["/test"]
        assert callable(router1.routes["/test"]["GET"]["middleware"])

    def test_include_router_with_class_middleware(self):
        async def class_middleware(endpoint, *args, **kwargs):
            return await endpoint(*args, **kwargs)

        async def handler(*args, **kwargs):
            return "handler executed"

        router1 = RouterBase(middleware=class_middleware)
        router2 = RouterBase()
        router2.add_route("/test", "GET", handler)

        router1.include_router(router2)

        assert "/test" in router1.routes
        assert "GET" in router1.routes["/test"]
        assert callable(router1.routes["/test"]["GET"]["middleware"])

    def test_include_router_with_only_class_middleware(self):
        async def class_middleware(endpoint, *args, **kwargs):
            return await endpoint(*args, **kwargs)

        async def handler(*args, **kwargs):
            return "handler executed"

        router1 = RouterBase(middleware=class_middleware)
        router2 = RouterBase()
        router2.add_route("/test", "GET", handler)

        router1.include_router(router2)

        assert "/test" in router1.routes
        assert "GET" in router1.routes["/test"]
        assert callable(router1.routes["/test"]["GET"]["middleware"])

    def test_include_router_with_only_method_middleware(self):
        async def method_middleware(endpoint, *args, **kwargs):
            return await endpoint(*args, **kwargs)

        async def handler(*args, **kwargs):
            return "handler executed"

        router1 = RouterBase()
        router2 = RouterBase()
        router2.add_route("/test", "GET", handler, middleware=method_middleware)

        router1.include_router(router2)

        assert "/test" in router1.routes
        assert "GET" in router1.routes["/test"]
        assert callable(router1.routes["/test"]["GET"]["middleware"])

    def test_include_endpoint_class_with_add_route(self):
        class MockEndpoint(Endpoint):
            path = "/mock"

            @Endpoint.GET()
            def get(self):
                return 200, {"message": "GET /mock"}

        router = RouterBase()
        router.include_endpoint_class(MockEndpoint)

        assert "/mock" in router.routes
        assert "GET" in router.routes["/mock"]
        assert callable(router.routes["/mock"]["GET"]["handler"])


class WSEndpoint(Endpoint):
    path = "/ws/chat"

    @Endpoint.MIDDLEWARE()
    async def ws_middleware(self, endpoint, headers, **kwargs):
        logger.info("middleware: Before to call the endpoint")
        yield await endpoint(headers=headers, **kwargs)
        logger.info("middleware: After to call the endpoint")

    async def connection(self, headers):
        logger.info("Client connected")
        return True

    async def disconnect(self, connection_id, headers):
        logger.info(f"Client {connection_id} disconnected")

    async def ws(self, connection_id, message, headers):
        logger.info(f"Received message from {connection_id}: {message}")
        for i in range(10):
            yield f"event: message\ndata: {i}\n\n"
            await asyncio.sleep(0.2)


class TestEndpoint:
    class MockEndpoint(Endpoint):
        path = "/mock"

        @Endpoint.GET()
        def get(self):
            return 200, {"message": "GET /mock"}

        @Endpoint.POST()
        def post(self):
            return 201, {"message": "POST /mock"}

        @Endpoint.PUT()
        def put(self):
            return 200, {"message": "PUT /mock"}

        @Endpoint.DELETE()
        def delete(self):
            return 200, {"message": "DELETE /mock"}

        @Endpoint.PATCH()
        def patch(self):
            return 200, {"message": "PATCH /mock"}

    def test_all_routes(self):
        endpoint = self.MockEndpoint()
        routes = endpoint.get_routes()
        assert "/mock" in routes
        assert "GET" in routes["/mock"]
        assert "POST" in routes["/mock"]
        assert "PUT" in routes["/mock"]
        assert "DELETE" in routes["/mock"]
        assert "PATCH" in routes["/mock"]

    def test_get_routes(self):
        endpoint = self.MockEndpoint()
        routes = endpoint.get_routes()
        assert "/mock" in routes
        assert "GET" in routes["/mock"]
        assert callable(routes["/mock"]["GET"]["handler"])

    def test_post_routes(self):
        endpoint = self.MockEndpoint()
        routes = endpoint.get_routes()
        assert "/mock" in routes
        assert "POST" in routes["/mock"]
        assert callable(routes["/mock"]["POST"]["handler"])

    def test_put_routes(self):
        endpoint = self.MockEndpoint()
        routes = endpoint.get_routes()
        assert "/mock" in routes
        assert "PUT" in routes["/mock"]
        assert callable(routes["/mock"]["PUT"]["handler"])

    def test_delete_routes(self):
        endpoint = self.MockEndpoint()
        routes = endpoint.get_routes()
        assert "/mock" in routes
        assert "DELETE" in routes["/mock"]
        assert callable(routes["/mock"]["DELETE"]["handler"])

    def test_patch_routes(self):
        endpoint = self.MockEndpoint()
        routes = endpoint.get_routes()
        assert "/mock" in routes
        assert "PATCH" in routes["/mock"]
        assert callable(routes["/mock"]["PATCH"]["handler"])

    def test_get_routes_with_annotations(self):
        class MockEndpointWithAnnotations(Endpoint):
            path = "/annotated"

            @Endpoint.GET()
            def get(self, param: PathParam[int]):
                return 200, {"message": f"GET /annotated with param {param}"}

        endpoint = MockEndpointWithAnnotations()
        routes = endpoint.get_routes()
        assert "/annotated/<param>" in routes
        assert "GET" in routes["/annotated/<param>"]
        assert callable(routes["/annotated/<param>"]["GET"]["handler"])

    def test_get_routes_without_param_names(self):
        class MockEndpointWithoutParams(Endpoint):
            path = "/noparams"

            @Endpoint.GET()
            def get(self):
                return 200, {"message": "GET /noparams"}

        endpoint = MockEndpointWithoutParams()
        routes = endpoint.get_routes()
        assert "/noparams" in routes
        assert "GET" in routes["/noparams"]
        assert callable(routes["/noparams"]["GET"]["handler"])
        assert routes["/noparams"]["GET"]["handler"]() == (
            200,
            {"message": "GET /noparams"},
        )

    def test_get_routes_with_param_names(self):
        class MockEndpointWithParams(Endpoint):
            path = "/params"

            @Endpoint.GET()
            def get(self, db: str):
                return 200, {"message": "GET /params without param"}

        endpoint = MockEndpointWithParams()
        routes = endpoint.get_routes()
        assert "/params" in routes
        assert "GET" in routes["/params"]
        assert callable(routes["/params"]["GET"]["handler"])

    def test_get_routes_with_both_middlewares(self):
        class MockEndpointWithMiddlewares(Endpoint):
            path = "/middlewares"

            @Endpoint.MIDDLEWARE()
            async def class_middleware(self, endpoint, *args, **kwargs):
                return await endpoint(*args, **kwargs)

            @Endpoint.GET(
                middleware=lambda endpoint, *args, **kwargs: endpoint(*args, **kwargs)
            )
            def get(self):
                return 200, {"message": "GET /middlewares"}

        endpoint = MockEndpointWithMiddlewares()
        routes = endpoint.get_routes()
        assert "/middlewares" in routes
        assert "GET" in routes["/middlewares"]
        assert callable(routes["/middlewares"]["GET"]["middleware"])

    def test_get_routes_with_class_middleware(self):
        class MockEndpointWithClassMiddleware(Endpoint):
            path = "/classmiddleware"

            @Endpoint.MIDDLEWARE()
            async def class_middleware(self, endpoint, *args, **kwargs):
                return await endpoint(*args, **kwargs)

            @Endpoint.GET()
            def get(self):
                return 200, {"message": "GET /classmiddleware"}

        endpoint = MockEndpointWithClassMiddleware()
        routes = endpoint.get_routes()
        assert "/classmiddleware" in routes
        assert "GET" in routes["/classmiddleware"]
        assert callable(routes["/classmiddleware"]["GET"]["middleware"])

    def test_get_routes_with_method_middleware(self):
        class MockEndpointWithMethodMiddleware(Endpoint):
            path = "/methodmiddleware"

            @Endpoint.GET(
                middleware=lambda endpoint, *args, **kwargs: endpoint(*args, **kwargs)
            )
            def get(self):
                return 200, {"message": "GET /methodmiddleware"}

        endpoint = MockEndpointWithMethodMiddleware()
        routes = endpoint.get_routes()
        assert "/methodmiddleware" in routes
        assert "GET" in routes["/methodmiddleware"]
        assert callable(routes["/methodmiddleware"]["GET"]["middleware"])


class TestBaseEndpoint:
    def test_middleware_decorator(self):
        class MockEndpoint(Endpoint):
            @Endpoint.MIDDLEWARE()
            def custom_middleware(cls, *args, **kwargs):
                return "middleware executed"

        endpoint = MockEndpoint()
        endpoint.get_routes()
        assert endpoint.middleware is not None
        assert endpoint.middleware() == "middleware executed"
